import { Injectable, NgZone, InjectionToken, Directive, ElementRef, Renderer2, Inject, Input, IterableDiffers, EventEmitter, Output, NgModule } from '@angular/core';
import { __awaiter } from 'tslib';
import { BehaviorSubject, Subject, of } from 'rxjs';
import { share } from 'rxjs/operators';
import { NgModel } from '@angular/forms';

/**
 * @fileoverview added by tsickle
 * Generated from: lib/util/ngx-validator-rules.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxValidatorRules {
    constructor() {
        this.isRequired = (/**
         * @template THIS
         * @this {THIS}
         * @param {?} errorMsg
         * @return {THIS}
         */
        (errorMsg) => {
            this.validationFns.push(this._isRequiredCheck.bind(this, errorMsg));
            return this;
        });
        this.isNumberWithinRange = (/**
         * @template THIS
         * @this {THIS}
         * @param {?} errorMsg
         * @param {?} range
         * @return {THIS}
         */
        (errorMsg, range) => {
            this.validationFns.push(this._isNumberWithinRangeCheck.bind(this, errorMsg, range));
            return this;
        });
        this.isNumber = (/**
         * @template THIS
         * @this {THIS}
         * @param {?} errorMsg
         * @return {THIS}
         */
        (errorMsg) => {
            this.validationFns.push(this._isNumberCheck.bind(this, errorMsg));
            return this;
        });
        this.isRegex = (/**
         * @template THIS
         * @this {THIS}
         * @param {?} errorMsg
         * @param {?} regex
         * @return {THIS}
         */
        (errorMsg, regex) => {
            this.validationFns.push(this._isRegexCheck.bind(this, errorMsg, regex));
            return this;
        });
        this.customValidator = (/**
         * @template THIS
         * @this {THIS}
         * @param {?} errorMsg
         * @param {?} customFunction
         * @return {THIS}
         */
        (errorMsg, customFunction) => {
            if (typeof (customFunction) !== "function") {
                throw new Error(`Invalid custom function. Received ${typeof (customFunction)}`);
            }
            this.validationFns.push(customFunction.bind(this, errorMsg));
            return this;
        });
        //////////////////////////
        // private helper function
        //////////////////////////
        this.isNullOrEmpty = (/**
         * @param {?} value
         * @return {?}
         */
        (value) => {
            // Allow 0 and false as a valid input
            if (!value && value !== 0 && value !== false) {
                return true;
            }
            return false;
        });
        this._isRequiredCheck = (/**
         * @param {?} errorMsg
         * @param {?=} value
         * @return {?}
         */
        (errorMsg, value) => {
            if (this.isNullOrEmpty(value)) {
                this.errorMsg = errorMsg || 'This value is required';
            }
            else {
                this.errorMsg = null;
            }
        });
        this._isNumberCheck = (/**
         * @param {?} errorMsg
         * @param {?=} value
         * @return {?}
         */
        (errorMsg, value) => {
            if (isNaN(parseFloat(value))) {
                this.errorMsg = errorMsg || 'It should be number';
            }
            else {
                this.errorMsg = null;
            }
        });
        this._isNumberWithinRangeCheck = (/**
         * @param {?} errorMsg
         * @param {?} range
         * @param {?=} value
         * @return {?}
         */
        (errorMsg, range, value) => {
            if (isNaN(parseFloat(String(value)))) {
                this.errorMsg = 'It should be number';
            }
            else if (!(parseFloat(value) >= range[0] && parseFloat(value) <= range[1])) {
                this.errorMsg = errorMsg || 'It should be within range';
            }
            else {
                this.errorMsg = null;
            }
        });
        this._isRegexCheck = (/**
         * @param {?} errorMsg
         * @param {?} regex
         * @param {?=} value
         * @return {?}
         */
        (errorMsg, regex, value) => {
            if (!regex.test(value)) {
                this.errorMsg = errorMsg || 'Invalid value';
            }
            else {
                this.errorMsg = null;
            }
        });
        this.validationFns = [];
    }
}
if (false) {
    /** @type {?} */
    NgxValidatorRules.prototype.propertyName;
    /** @type {?} */
    NgxValidatorRules.prototype.dependency;
    /** @type {?} */
    NgxValidatorRules.prototype.errorMsg;
    /** @type {?} */
    NgxValidatorRules.prototype.validationFns;
    /** @type {?} */
    NgxValidatorRules.prototype.isRequired;
    /** @type {?} */
    NgxValidatorRules.prototype.isNumberWithinRange;
    /** @type {?} */
    NgxValidatorRules.prototype.isNumber;
    /** @type {?} */
    NgxValidatorRules.prototype.isRegex;
    /** @type {?} */
    NgxValidatorRules.prototype.customValidator;
    /**
     * @type {?}
     * @private
     */
    NgxValidatorRules.prototype.isNullOrEmpty;
    /**
     * @type {?}
     * @private
     */
    NgxValidatorRules.prototype._isRequiredCheck;
    /**
     * @type {?}
     * @private
     */
    NgxValidatorRules.prototype._isNumberCheck;
    /**
     * @type {?}
     * @private
     */
    NgxValidatorRules.prototype._isNumberWithinRangeCheck;
    /**
     * @type {?}
     * @private
     */
    NgxValidatorRules.prototype._isRegexCheck;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/util/ngx-validator-helper.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxValidatorHelper {
    constructor() {
        this.validateFor = (/**
         * @param {?} propName
         * @param {?=} dependency
         * @return {?}
         */
        (propName, dependency) => {
            /** @type {?} */
            const ngxValidatorRules = new NgxValidatorRules();
            ngxValidatorRules.propertyName = propName;
            ngxValidatorRules.dependency = dependency;
            return ngxValidatorRules;
        });
    }
}
if (false) {
    /** @type {?} */
    NgxValidatorHelper.prototype.validateFor;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/expression_parser/parse.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// Copied from https://github.com/peerigon/angular-expressions/blob/master/lib/parse.js
"use strict";
/** @type {?} */
var _window_ = window || { document: {} };
////////////////////////////////////
/**
 * \@ngdoc module
 * \@name ng
 * \@module ng
 * \@installation
 * \@description
 *
 * # ng (core module)
 * The ng module is loaded by default when an AngularJS application is started. The module itself
 * contains the essential components for an AngularJS application to function. The table below
 * lists a high level breakdown of each of the services/factories, filters, directives and testing
 * components available within this core module.
 *
 * <div doc-module-components="ng"></div>
 * @type {?}
 */
var NODE_TYPE_ELEMENT = 1;
/** @type {?} */
var NODE_TYPE_ATTRIBUTE = 2;
/** @type {?} */
var NODE_TYPE_TEXT = 3;
/** @type {?} */
var NODE_TYPE_COMMENT = 8;
/** @type {?} */
var NODE_TYPE_DOCUMENT = 9;
/** @type {?} */
var NODE_TYPE_DOCUMENT_FRAGMENT = 11;
/** @type {?} */
var REGEX_STRING_REGEXP = /^\/(.+)\/([a-z]*)$/;
// The name of a form control's ValidityState property.
// This is used so that it's possible for internal tests to create mock ValidityStates.
/** @type {?} */
var VALIDITY_STATE_PROPERTY = "validity";
/** @type {?} */
var hasOwnProperty = Object.prototype.hasOwnProperty;
/** @type {?} */
var lowercase = (/**
 * @param {?} string
 * @return {?}
 */
function (string) {
    return isString(string) ? string.toLowerCase() : string;
});
const ɵ0 = lowercase;
/** @type {?} */
var uppercase = (/**
 * @param {?} string
 * @return {?}
 */
function (string) {
    return isString(string) ? string.toUpperCase() : string;
});
const ɵ1 = uppercase;
/**
 * \@ngdoc function
 * \@name angular.isArray
 * \@module ng
 * \@kind function
 *
 * \@description
 * Determines if a reference is an `Array`.
 *
 * @type {?}
 */
var isArray = Array.isArray;
/** @type {?} */
var ngAttrPrefixes = ["ng-", "data-ng-", "ng:", "x-ng-"];
/** @type {?} */
var manualLowercase = (/**
 * @param {?} s
 * @return {?}
 */
function (s) {
    /* eslint-disable no-bitwise */
    return isString(s)
        ? s.replace(/[A-Z]/g, (/**
         * @param {?} ch
         * @return {?}
         */
        function (ch) {
            return String.fromCharCode(ch.charCodeAt(0) | 32);
        }))
        : s;
    /* eslint-enable */
});
const ɵ2 = manualLowercase;
/** @type {?} */
var manualUppercase = (/**
 * @param {?} s
 * @return {?}
 */
function (s) {
    /* eslint-disable no-bitwise */
    return isString(s)
        ? s.replace(/[a-z]/g, (/**
         * @param {?} ch
         * @return {?}
         */
        function (ch) {
            return String.fromCharCode(ch.charCodeAt(0) & ~32);
        }))
        : s;
    /* eslint-enable */
});
const ɵ3 = manualUppercase;
// String#toLowerCase and String#toUpperCase don't produce correct results in browsers with Turkish
// locale, for this reason we need to detect this case and redefine lowercase/uppercase methods
// with correct but slower alternatives. See https://github.com/angular/angular.js/issues/11387
if ("i" !== "I".toLowerCase()) {
    lowercase = manualLowercase;
    uppercase = manualUppercase;
}
/** @type {?} */
var msie;
/** @type {?} */
var // holds major version number for IE, or NaN if UA is not IE.
jqLite;
/** @type {?} */
var // delay binding since jQuery could be loaded after us.
jQuery;
/** @type {?} */
var // delay binding
slice = [].slice;
/** @type {?} */
var splice = [].splice;
/** @type {?} */
var push = [].push;
/** @type {?} */
var 
// toString = Object.prototype.toString,
getPrototypeOf = Object.getPrototypeOf;
/** @type {?} */
var 
// ngMinErr = minErr("ng"),
uid = 0;
/**
 * \@ngdoc function
 * \@name angular.noop
 * \@module ng
 * \@kind function
 *
 * \@description
 * A function that performs no operations. This function can be useful when writing code in the
 * functional style.
 * ```js
 * function foo(callback) {
 * var result = calculateResult();
 * (callback || angular.noop)(result);
 * }
 * ```
 * @return {?}
 */
function noop() { }
/**
 * \@ngdoc function
 * \@name angular.isUndefined
 * \@module ng
 * \@kind function
 *
 * \@description
 * Determines if a reference is undefined.
 *
 * param {*} value Reference to check.
 * returns {boolean} True if `value` is undefined.
 * @param {?} value
 * @return {?}
 */
function isUndefined(value) {
    return typeof value === "undefined";
}
/**
 * \@ngdoc function
 * \@name angular.isDefined
 * \@module ng
 * \@kind function
 *
 * \@description
 * Determines if a reference is defined.
 *
 * param {*} value Reference to check.
 * returns {boolean} True if `value` is defined.
 * @param {?} value
 * @return {?}
 */
function isDefined(value) {
    return typeof value !== "undefined";
}
/**
 * \@ngdoc function
 * \@name angular.isObject
 * \@module ng
 * \@kind function
 *
 * \@description
 * Determines if a reference is an `Object`. Unlike `typeof` in JavaScript, `null`s are not
 * considered to be objects. Note that JavaScript arrays are objects.
 *
 * param {*} value Reference to check.
 * returns {boolean} True if `value` is an `Object` but not `null`.
 * @param {?} value
 * @return {?}
 */
function isObject(value) {
    // http://jsperf.com/isobject4
    return value !== null && typeof value === "object";
}
/**
 * Determine if a value is an object with a null prototype
 *
 * returns {boolean} True if `value` is an `Object` with a null prototype
 * @param {?} value
 * @return {?}
 */
function isBlankObject(value) {
    return value !== null && typeof value === "object" && !getPrototypeOf(value);
}
/**
 * \@ngdoc function
 * \@name angular.isString
 * \@module ng
 * \@kind function
 *
 * \@description
 * Determines if a reference is a `String`.
 *
 * param {*} value Reference to check.
 * returns {boolean} True if `value` is a `String`.
 * @param {?} value
 * @return {?}
 */
function isString(value) {
    return typeof value === "string";
}
/**
 * \@ngdoc function
 * \@name angular.isNumber
 * \@module ng
 * \@kind function
 *
 * \@description
 * Determines if a reference is a `Number`.
 *
 * This includes the "special" numbers `NaN`, `+Infinity` and `-Infinity`.
 *
 * If you wish to exclude these then you can use the native
 * [`isFinite'](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/isFinite)
 * method.
 *
 * param {*} value Reference to check.
 * returns {boolean} True if `value` is a `Number`.
 * @param {?} value
 * @return {?}
 */
function isNumber(value) {
    return typeof value === "number";
}
/**
 * \@ngdoc function
 * \@name angular.isFunction
 * \@module ng
 * \@kind function
 *
 * \@description
 * Determines if a reference is a `Function`.
 *
 * param {*} value Reference to check.
 * returns {boolean} True if `value` is a `Function`.
 * @param {?} value
 * @return {?}
 */
function isFunction(value) {
    return typeof value === "function";
}
/**
 * Checks if `obj` is a window object.
 *
 * private
 * param {*} obj Object to check
 * returns {boolean} True if `obj` is a window obj.
 * @param {?} obj
 * @return {?}
 */
function isWindow(obj) {
    return obj && obj.window === obj;
}
/**
 * @param {?} obj
 * @return {?}
 */
function isScope(obj) {
    return obj && obj.$evalAsync && obj.$watch;
}
/**
 * @param {?} key
 * @param {?} value
 * @return {?}
 */
function toJsonReplacer(key, value) {
    /** @type {?} */
    var val = value;
    if (typeof key === "string" &&
        key.charAt(0) === "$" &&
        key.charAt(1) === "$") {
        val = undefined;
    }
    else if (isWindow(value)) {
        val = "$WINDOW";
    }
    else if (value && window.document === value) {
        val = "$DOCUMENT";
    }
    else if (isScope(value)) {
        val = "$SCOPE";
    }
    return val;
}
/**
 * Creates a new object without a prototype. This object is useful for lookup without having to
 * guard against prototypically inherited properties via hasOwnProperty.
 *
 * Related micro-benchmarks:
 * - http://jsperf.com/object-create2
 * - http://jsperf.com/proto-map-lookup/2
 * - http://jsperf.com/for-in-vs-object-keys2
 *
 * returns {Object}
 * @return {?}
 */
function createMap() {
    return Object.create(null);
}
/**
 * @param {?} value
 * @return {?}
 */
function stringify(value) {
    return JSON.stringify(value);
}
/* global toDebugString: true */
/**
 * @param {?} obj
 * @return {?}
 */
function serializeObject(obj) {
    /** @type {?} */
    var seen = [];
    return JSON.stringify(obj, (/**
     * @param {?} key
     * @param {?} val
     * @return {?}
     */
    function (key, val) {
        val = toJsonReplacer(key, val);
        if (isObject(val)) {
            if (seen.indexOf(val) >= 0)
                return "...";
            seen.push(val);
        }
        return val;
    }));
}
/**
 * @param {?} obj
 * @return {?}
 */
function toDebugString(obj) {
    if (typeof obj === "function") {
        return obj.toString().replace(/ \{[\s\S]*$/, "");
    }
    else if (isUndefined(obj)) {
        return "undefined";
    }
    else if (typeof obj !== "string") {
        return serializeObject(obj);
    }
    return obj;
}
/** @type {?} */
var objectValueOf = {}.constructor.prototype.valueOf;
// Sandboxing Angular Expressions
// ------------------------------
// Angular expressions are no longer sandboxed. So it is now even easier to access arbitrary JS code by
// various means such as obtaining a reference to native JS functions like the Function constructor.
//
// As an example, consider the following Angular expression:
//
//   {}.toString.constructor('alert("evil JS code")')
//
// It is important to realize that if you create an expression from a string that contains user provided
// content then it is possible that your application contains a security vulnerability to an XSS style attack.
//
// See https://docs.angularjs.org/guide/security
/**
 * @param {?} name
 * @return {?}
 */
function getStringValue(name) {
    // Property names must be strings. This means that non-string objects cannot be used
    // as keys in an object. Any non-string object, including a number, is typecasted
    // into a string via the toString method.
    // -- MDN, https://developer.mozilla.org/en/docs/Web/JavaScript/Reference/Operators/Property_accessors#Property_names
    //
    // So, to ensure that we are checking the same `name` that JavaScript would use, we cast it
    // to a string. It's not always possible. If `name` is an object and its `toString` method is
    // 'broken' (doesn't return a string, isn't a function, etc.), an error will be thrown:
    //
    // TypeError: Cannot convert object to primitive value
    //
    // For performance reasons, we don't catch this error here and allow it to propagate up the call
    // stack. Note that you'll get the same error in JavaScript if you try to access a property using
    // such a 'broken' object as a key.
    return name + "";
}
/** @type {?} */
var OPERATORS = createMap();
"+ - * / % === !== == != < > <= >= && || ! = |".split(" ").forEach((/**
 * @param {?} operator
 * @return {?}
 */
function (operator) {
    OPERATORS[operator] = true;
}));
/** @type {?} */
var ESCAPE = {
    n: "\n",
    f: "\f",
    r: "\r",
    t: "\t",
    v: "\v",
    "'": "'",
    '"': '"'
};
/////////////////////////////////////////
/**
 * constructor
 * @param {?} options
 * @return {?}
 */
function Lexer(options) {
    this.options = options;
}
;
Lexer.prototype = {
    constructor: Lexer,
    lex: (/**
     * @param {?} text
     * @return {?}
     */
    function (text) {
        this.text = text;
        this.index = 0;
        this.tokens = [];
        while (this.index < this.text.length) {
            /** @type {?} */
            var ch = this.text.charAt(this.index);
            if (ch === '"' || ch === "'") {
                this.readString(ch);
            }
            else if (this.isNumber(ch) ||
                (ch === "." && this.isNumber(this.peek()))) {
                this.readNumber();
            }
            else if (this.isIdentifierStart(this.peekMultichar())) {
                this.readIdent();
            }
            else if (this.is(ch, "(){}[].,;:?")) {
                this.tokens.push({ index: this.index, text: ch });
                this.index++;
            }
            else if (this.isWhitespace(ch)) {
                this.index++;
            }
            else {
                /** @type {?} */
                var ch2 = ch + this.peek();
                /** @type {?} */
                var ch3 = ch2 + this.peek(2);
                /** @type {?} */
                var op1 = OPERATORS[ch];
                /** @type {?} */
                var op2 = OPERATORS[ch2];
                /** @type {?} */
                var op3 = OPERATORS[ch3];
                if (op1 || op2 || op3) {
                    /** @type {?} */
                    var token = op3 ? ch3 : op2 ? ch2 : ch;
                    this.tokens.push({ index: this.index, text: token, operator: true });
                    this.index += token.length;
                }
                else {
                    this.throwError("Unexpected next character ", this.index, this.index + 1);
                }
            }
        }
        return this.tokens;
    }),
    is: (/**
     * @param {?} ch
     * @param {?} chars
     * @return {?}
     */
    function (ch, chars) {
        return chars.indexOf(ch) !== -1;
    }),
    peek: (/**
     * @param {?} i
     * @return {?}
     */
    function (i) {
        /** @type {?} */
        var num = i || 1;
        return this.index + num < this.text.length
            ? this.text.charAt(this.index + num)
            : false;
    }),
    isNumber: (/**
     * @param {?} ch
     * @return {?}
     */
    function (ch) {
        return "0" <= ch && ch <= "9" && typeof ch === "string";
    }),
    isWhitespace: (/**
     * @param {?} ch
     * @return {?}
     */
    function (ch) {
        // IE treats non-breaking space as \u00A0
        return (ch === " " ||
            ch === "\r" ||
            ch === "\t" ||
            ch === "\n" ||
            ch === "\v" ||
            ch === "\u00A0");
    }),
    isIdentifierStart: (/**
     * @param {?} ch
     * @return {?}
     */
    function (ch) {
        return this.options.isIdentifierStart
            ? this.options.isIdentifierStart(ch, this.codePointAt(ch))
            : this.isValidIdentifierStart(ch);
    }),
    isValidIdentifierStart: (/**
     * @param {?} ch
     * @return {?}
     */
    function (ch) {
        return (("a" <= ch && ch <= "z") ||
            ("A" <= ch && ch <= "Z") ||
            "_" === ch ||
            ch === "$");
    }),
    isIdentifierContinue: (/**
     * @param {?} ch
     * @return {?}
     */
    function (ch) {
        return this.options.isIdentifierContinue
            ? this.options.isIdentifierContinue(ch, this.codePointAt(ch))
            : this.isValidIdentifierContinue(ch);
    }),
    isValidIdentifierContinue: (/**
     * @param {?} ch
     * @param {?} cp
     * @return {?}
     */
    function (ch, cp) {
        return this.isValidIdentifierStart(ch, cp) || this.isNumber(ch);
    }),
    codePointAt: (/**
     * @param {?} ch
     * @return {?}
     */
    function (ch) {
        if (ch.length === 1)
            return ch.charCodeAt(0);
        // eslint-disable-next-line no-bitwise
        return (ch.charCodeAt(0) << 10) + ch.charCodeAt(1) - 0x35fdc00;
    }),
    peekMultichar: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var ch = this.text.charAt(this.index);
        /** @type {?} */
        var peek = this.peek();
        if (!peek) {
            return ch;
        }
        /** @type {?} */
        var cp1 = ch.charCodeAt(0);
        /** @type {?} */
        var cp2 = peek.charCodeAt(0);
        if (cp1 >= 0xd800 && cp1 <= 0xdbff && cp2 >= 0xdc00 && cp2 <= 0xdfff) {
            return ch + peek;
        }
        return ch;
    }),
    isExpOperator: (/**
     * @param {?} ch
     * @return {?}
     */
    function (ch) {
        return ch === "-" || ch === "+" || this.isNumber(ch);
    }),
    throwError: (/**
     * @param {?} error
     * @param {?} start
     * @param {?} end
     * @return {?}
     */
    function (error, start, end) {
        end = end || this.index;
        /** @type {?} */
        var colStr = isDefined(start)
            ? "s " +
                start +
                "-" +
                this.index +
                " [" +
                this.text.substring(start, end) +
                "]"
            : " " + end;
        throw this.throwError("lexerr", "Lexer Error: {0} at column{1} in expression [{2}].", error, colStr, this.text);
    }),
    readNumber: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var number = "";
        /** @type {?} */
        var start = this.index;
        while (this.index < this.text.length) {
            /** @type {?} */
            var ch = lowercase(this.text.charAt(this.index));
            if (ch === "." || this.isNumber(ch)) {
                number += ch;
            }
            else {
                /** @type {?} */
                var peekCh = this.peek();
                if (ch === "e" && this.isExpOperator(peekCh)) {
                    number += ch;
                }
                else if (this.isExpOperator(ch) &&
                    peekCh &&
                    this.isNumber(peekCh) &&
                    number.charAt(number.length - 1) === "e") {
                    number += ch;
                }
                else if (this.isExpOperator(ch) &&
                    (!peekCh || !this.isNumber(peekCh)) &&
                    number.charAt(number.length - 1) === "e") {
                    this.throwError("Invalid exponent");
                }
                else {
                    break;
                }
            }
            this.index++;
        }
        this.tokens.push({
            index: start,
            text: number,
            constant: true,
            value: Number(number)
        });
    }),
    readIdent: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var start = this.index;
        this.index += this.peekMultichar().length;
        while (this.index < this.text.length) {
            /** @type {?} */
            var ch = this.peekMultichar();
            if (!this.isIdentifierContinue(ch)) {
                break;
            }
            this.index += ch.length;
        }
        this.tokens.push({
            index: start,
            text: this.text.slice(start, this.index),
            identifier: true
        });
    }),
    readString: (/**
     * @param {?} quote
     * @return {?}
     */
    function (quote) {
        /** @type {?} */
        var start = this.index;
        this.index++;
        /** @type {?} */
        var string = "";
        /** @type {?} */
        var rawString = quote;
        /** @type {?} */
        var escape = false;
        while (this.index < this.text.length) {
            /** @type {?} */
            var ch = this.text.charAt(this.index);
            rawString += ch;
            if (escape) {
                if (ch === "u") {
                    /** @type {?} */
                    var hex = this.text.substring(this.index + 1, this.index + 5);
                    if (!hex.match(/[\da-f]{4}/i)) {
                        this.throwError("Invalid unicode escape [\\u" + hex + "]");
                    }
                    this.index += 4;
                    string += String.fromCharCode(parseInt(hex, 16));
                }
                else {
                    /** @type {?} */
                    var rep = ESCAPE[ch];
                    string = string + (rep || ch);
                }
                escape = false;
            }
            else if (ch === "\\") {
                escape = true;
            }
            else if (ch === quote) {
                this.index++;
                this.tokens.push({
                    index: start,
                    text: rawString,
                    constant: true,
                    value: string
                });
                return;
            }
            else {
                string += ch;
            }
            this.index++;
        }
        this.throwError("Unterminated quote", start);
    })
};
/** @type {?} */
var AST = (/**
 * @param {?} lexer
 * @param {?} options
 * @return {?}
 */
function AST(lexer, options) {
    this.lexer = lexer;
    this.options = options;
});
const ɵ4 = AST;
AST.prototype = {
    ast: (/**
     * @param {?} text
     * @return {?}
     */
    function (text) {
        this.text = text;
        this.tokens = this.lexer.lex(text);
        /** @type {?} */
        var value = this.program();
        if (this.tokens.length !== 0) {
            this.throwError("is an unexpected token", this.tokens[0]);
        }
        return value;
    }),
    program: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var body = [];
        while (true) {
            if (this.tokens.length > 0 && !this.peek("}", ")", ";", "]"))
                body.push(this.expressionStatement());
            if (!this.expect(";")) {
                return { type: AST.prototype.Program, body: body };
            }
        }
    }),
    expressionStatement: (/**
     * @return {?}
     */
    function () {
        return { type: AST.prototype.ExpressionStatement, expression: this.filterChain() };
    }),
    filterChain: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var left = this.expression();
        while (this.expect("|")) {
            left = this.filter(left);
        }
        return left;
    }),
    expression: (/**
     * @return {?}
     */
    function () {
        return this.assignment();
    }),
    assignment: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var result = this.ternary();
        if (this.expect("=")) {
            if (!isAssignable(result)) {
                throw this.throwError("lval", "Trying to assign a value to a non l-value");
            }
            result = {
                type: AST.prototype.AssignmentExpression,
                left: result,
                right: this.assignment(),
                operator: "="
            };
        }
        return result;
    }),
    ternary: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var test = this.logicalOR();
        /** @type {?} */
        var alternate;
        /** @type {?} */
        var consequent;
        if (this.expect("?")) {
            alternate = this.expression();
            if (this.consume(":")) {
                consequent = this.expression();
                return {
                    type: AST.prototype.ConditionalExpression,
                    test: test,
                    alternate: alternate,
                    consequent: consequent
                };
            }
        }
        return test;
    }),
    logicalOR: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var left = this.logicalAND();
        while (this.expect("||")) {
            left = {
                type: AST.prototype.LogicalExpression,
                operator: "||",
                left: left,
                right: this.logicalAND()
            };
        }
        return left;
    }),
    logicalAND: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var left = this.equality();
        while (this.expect("&&")) {
            left = {
                type: AST.prototype.LogicalExpression,
                operator: "&&",
                left: left,
                right: this.equality()
            };
        }
        return left;
    }),
    equality: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var left = this.relational();
        /** @type {?} */
        var token;
        while ((token = this.expect("==", "!=", "===", "!=="))) {
            left = {
                type: AST.prototype.BinaryExpression,
                operator: token.text,
                left: left,
                right: this.relational()
            };
        }
        return left;
    }),
    relational: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var left = this.additive();
        /** @type {?} */
        var token;
        while ((token = this.expect("<", ">", "<=", ">="))) {
            left = {
                type: AST.prototype.BinaryExpression,
                operator: token.text,
                left: left,
                right: this.additive()
            };
        }
        return left;
    }),
    additive: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var left = this.multiplicative();
        /** @type {?} */
        var token;
        while ((token = this.expect("+", "-"))) {
            left = {
                type: AST.prototype.BinaryExpression,
                operator: token.text,
                left: left,
                right: this.multiplicative()
            };
        }
        return left;
    }),
    multiplicative: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var left = this.unary();
        /** @type {?} */
        var token;
        while ((token = this.expect("*", "/", "%"))) {
            left = {
                type: AST.prototype.BinaryExpression,
                operator: token.text,
                left: left,
                right: this.unary()
            };
        }
        return left;
    }),
    unary: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var token;
        if ((token = this.expect("+", "-", "!"))) {
            return {
                type: AST.prototype.UnaryExpression,
                operator: token.text,
                prefix: true,
                argument: this.unary()
            };
        }
        else {
            return this.primary();
        }
    }),
    primary: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var primary;
        if (this.expect("(")) {
            primary = this.filterChain();
            this.consume(")");
        }
        else if (this.expect("[")) {
            primary = this.arrayDeclaration();
        }
        else if (this.expect("{")) {
            primary = this.object();
        }
        else if (this.selfReferential.hasOwnProperty(this.peek().text)) {
            primary = this.selfReferential[this.consume().text];
        }
        else if (this.options.literals.hasOwnProperty(this.peek().text)) {
            primary = {
                type: AST.prototype.Literal,
                value: this.options.literals[this.consume().text]
            };
        }
        else if (this.peek().identifier) {
            primary = this.identifier();
        }
        else if (this.peek().constant) {
            primary = this.constant();
        }
        else {
            this.throwError("not a primary expression", this.peek());
        }
        /** @type {?} */
        var next;
        while ((next = this.expect("(", "[", "."))) {
            if (next.text === "(") {
                primary = {
                    type: AST.prototype.CallExpression,
                    callee: primary,
                    arguments: this.parseArguments()
                };
                this.consume(")");
            }
            else if (next.text === "[") {
                primary = {
                    type: AST.prototype.MemberExpression,
                    object: primary,
                    property: this.expression(),
                    computed: true
                };
                this.consume("]");
            }
            else if (next.text === ".") {
                primary = {
                    type: AST.prototype.MemberExpression,
                    object: primary,
                    property: this.identifier(),
                    computed: false
                };
            }
            else {
                this.throwError("IMPOSSIBLE");
            }
        }
        return primary;
    }),
    filter: (/**
     * @param {?} baseExpression
     * @return {?}
     */
    function (baseExpression) {
        /** @type {?} */
        var args = [baseExpression];
        /** @type {?} */
        var result = {
            type: AST.prototype.CallExpression,
            callee: this.identifier(),
            arguments: args,
            filter: true
        };
        while (this.expect(":")) {
            args.push(this.expression());
        }
        return result;
    }),
    parseArguments: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var args = [];
        if (this.peekToken().text !== ")") {
            do {
                args.push(this.filterChain());
            } while (this.expect(","));
        }
        return args;
    }),
    identifier: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var token = this.consume();
        if (!token.identifier) {
            this.throwError("is not a valid identifier", token);
        }
        return { type: AST.prototype.Identifier, name: token.text };
    }),
    constant: (/**
     * @return {?}
     */
    function () {
        // TODO check that it is a constant
        return { type: AST.prototype.Literal, value: this.consume().value };
    }),
    arrayDeclaration: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var elements = [];
        if (this.peekToken().text !== "]") {
            do {
                if (this.peek("]")) {
                    // Support trailing commas per ES5.1.
                    break;
                }
                elements.push(this.expression());
            } while (this.expect(","));
        }
        this.consume("]");
        return { type: AST.prototype.ArrayExpression, elements: elements };
    }),
    object: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var properties = [];
        /** @type {?} */
        var property;
        if (this.peekToken().text !== "}") {
            do {
                if (this.peek("}")) {
                    // Support trailing commas per ES5.1.
                    break;
                }
                property = { type: AST.prototype.Property, kind: "init" };
                if (this.peek().constant) {
                    property.key = this.constant();
                    property.computed = false;
                    this.consume(":");
                    property.value = this.expression();
                }
                else if (this.peek().identifier) {
                    property.key = this.identifier();
                    property.computed = false;
                    if (this.peek(":")) {
                        this.consume(":");
                        property.value = this.expression();
                    }
                    else {
                        property.value = property.key;
                    }
                }
                else if (this.peek("[")) {
                    this.consume("[");
                    property.key = this.expression();
                    this.consume("]");
                    property.computed = true;
                    this.consume(":");
                    property.value = this.expression();
                }
                else {
                    this.throwError("invalid key", this.peek());
                }
                properties.push(property);
            } while (this.expect(","));
        }
        this.consume("}");
        return { type: AST.prototype.ObjectExpression, properties: properties };
    }),
    throwError: (/**
     * @param {?} msg
     * @param {?} token
     * @return {?}
     */
    function (msg, token) {
        throw this.throwError("syntax", "Syntax Error: Token '{0}' {1} at column {2} of the expression [{3}] starting at [{4}].", token.text, msg, token.index + 1, this.text, this.text.substring(token.index));
    }),
    consume: (/**
     * @param {?} e1
     * @return {?}
     */
    function (e1) {
        if (this.tokens.length === 0) {
            throw this.throwError("ueoe", "Unexpected end of expression: {0}", this.text);
        }
        /** @type {?} */
        var token = this.expect(e1);
        if (!token) {
            this.throwError("is unexpected, expecting [" + e1 + "]", this.peek());
        }
        return token;
    }),
    peekToken: (/**
     * @return {?}
     */
    function () {
        if (this.tokens.length === 0) {
            throw this.throwError("ueoe", "Unexpected end of expression: {0}", this.text);
        }
        return this.tokens[0];
    }),
    peek: (/**
     * @param {?} e1
     * @param {?} e2
     * @param {?} e3
     * @param {?} e4
     * @return {?}
     */
    function (e1, e2, e3, e4) {
        return this.peekAhead(0, e1, e2, e3, e4);
    }),
    peekAhead: (/**
     * @param {?} i
     * @param {?} e1
     * @param {?} e2
     * @param {?} e3
     * @param {?} e4
     * @return {?}
     */
    function (i, e1, e2, e3, e4) {
        if (this.tokens.length > i) {
            /** @type {?} */
            var token = this.tokens[i];
            /** @type {?} */
            var t = token.text;
            if (t === e1 ||
                t === e2 ||
                t === e3 ||
                t === e4 ||
                (!e1 && !e2 && !e3 && !e4)) {
                return token;
            }
        }
        return false;
    }),
    expect: (/**
     * @param {?} e1
     * @param {?} e2
     * @param {?} e3
     * @param {?} e4
     * @return {?}
     */
    function (e1, e2, e3, e4) {
        /** @type {?} */
        var token = this.peek(e1, e2, e3, e4);
        if (token) {
            this.tokens.shift();
            return token;
        }
        return false;
    }),
    selfReferential: {
        this: { type: AST.prototype.ThisExpression },
        $locals: { type: AST.prototype.LocalsExpression }
    }
};
AST.prototype.Program = "Program";
AST.prototype.ExpressionStatement = "ExpressionStatement";
AST.prototype.AssignmentExpression = "AssignmentExpression";
AST.prototype.ConditionalExpression = "ConditionalExpression";
AST.prototype.LogicalExpression = "LogicalExpression";
AST.prototype.BinaryExpression = "BinaryExpression";
AST.prototype.UnaryExpression = "UnaryExpression";
AST.prototype.CallExpression = "CallExpression";
AST.prototype.MemberExpression = "MemberExpression";
AST.prototype.Identifier = "Identifier";
AST.prototype.Literal = "Literal";
AST.prototype.ArrayExpression = "ArrayExpression";
AST.prototype.Property = "Property";
AST.prototype.ObjectExpression = "ObjectExpression";
AST.prototype.ThisExpression = "ThisExpression";
AST.prototype.LocalsExpression = "LocalsExpression";
// Internal use only
AST.prototype.NGValueParameter = "NGValueParameter";
/**
 * @param {?} v
 * @param {?} d
 * @return {?}
 */
function ifDefined(v, d) {
    return typeof v !== "undefined" ? v : d;
}
/**
 * @param {?} l
 * @param {?} r
 * @return {?}
 */
function plusFn(l, r) {
    if (typeof l === "undefined")
        return r;
    if (typeof r === "undefined")
        return l;
    return l + r;
}
/**
 * @param {?} $filter
 * @param {?} filterName
 * @return {?}
 */
function isStateless($filter, filterName) {
    /** @type {?} */
    var fn = $filter(filterName);
    return !fn.$stateful;
}
/**
 * @param {?} ast
 * @param {?} $filter
 * @return {?}
 */
function findConstantAndWatchExpressions(ast, $filter) {
    /** @type {?} */
    var allConstants;
    /** @type {?} */
    var argsToWatch;
    /** @type {?} */
    var isStatelessFilter;
    switch (ast.type) {
        case AST.prototype.Program:
            allConstants = true;
            ast.body.forEach((/**
             * @param {?} expr
             * @return {?}
             */
            function (expr) {
                findConstantAndWatchExpressions(expr.expression, $filter);
                allConstants = allConstants && expr.expression.constant;
            }));
            ast.constant = allConstants;
            break;
        case AST.prototype.Literal:
            ast.constant = true;
            ast.toWatch = [];
            break;
        case AST.prototype.UnaryExpression:
            findConstantAndWatchExpressions(ast.argument, $filter);
            ast.constant = ast.argument.constant;
            ast.toWatch = ast.argument.toWatch;
            break;
        case AST.prototype.BinaryExpression:
            findConstantAndWatchExpressions(ast.left, $filter);
            findConstantAndWatchExpressions(ast.right, $filter);
            ast.constant = ast.left.constant && ast.right.constant;
            ast.toWatch = ast.left.toWatch.concat(ast.right.toWatch);
            break;
        case AST.prototype.LogicalExpression:
            findConstantAndWatchExpressions(ast.left, $filter);
            findConstantAndWatchExpressions(ast.right, $filter);
            ast.constant = ast.left.constant && ast.right.constant;
            ast.toWatch = ast.constant ? [] : [ast];
            break;
        case AST.prototype.ConditionalExpression:
            findConstantAndWatchExpressions(ast.test, $filter);
            findConstantAndWatchExpressions(ast.alternate, $filter);
            findConstantAndWatchExpressions(ast.consequent, $filter);
            ast.constant =
                ast.test.constant && ast.alternate.constant && ast.consequent.constant;
            ast.toWatch = ast.constant ? [] : [ast];
            break;
        case AST.prototype.Identifier:
            ast.constant = false;
            ast.toWatch = [ast];
            break;
        case AST.prototype.MemberExpression:
            findConstantAndWatchExpressions(ast.object, $filter);
            if (ast.computed) {
                findConstantAndWatchExpressions(ast.property, $filter);
            }
            ast.constant =
                ast.object.constant && (!ast.computed || ast.property.constant);
            ast.toWatch = [ast];
            break;
        case AST.prototype.CallExpression:
            isStatelessFilter = ast.filter
                ? isStateless($filter, ast.callee.name)
                : false;
            allConstants = isStatelessFilter;
            argsToWatch = [];
            ast.arguments.forEach((/**
             * @param {?} expr
             * @return {?}
             */
            function (expr) {
                findConstantAndWatchExpressions(expr, $filter);
                allConstants = allConstants && expr.constant;
                if (!expr.constant) {
                    argsToWatch.push.apply(argsToWatch, expr.toWatch);
                }
            }));
            ast.constant = allConstants;
            ast.toWatch = isStatelessFilter ? argsToWatch : [ast];
            break;
        case AST.prototype.AssignmentExpression:
            findConstantAndWatchExpressions(ast.left, $filter);
            findConstantAndWatchExpressions(ast.right, $filter);
            ast.constant = ast.left.constant && ast.right.constant;
            ast.toWatch = [ast];
            break;
        case AST.prototype.ArrayExpression:
            allConstants = true;
            argsToWatch = [];
            ast.elements.forEach((/**
             * @param {?} expr
             * @return {?}
             */
            function (expr) {
                findConstantAndWatchExpressions(expr, $filter);
                allConstants = allConstants && expr.constant;
                if (!expr.constant) {
                    argsToWatch.push.apply(argsToWatch, expr.toWatch);
                }
            }));
            ast.constant = allConstants;
            ast.toWatch = argsToWatch;
            break;
        case AST.prototype.ObjectExpression:
            allConstants = true;
            argsToWatch = [];
            ast.properties.forEach((/**
             * @param {?} property
             * @return {?}
             */
            function (property) {
                findConstantAndWatchExpressions(property.value, $filter);
                allConstants =
                    allConstants && property.value.constant && !property.computed;
                if (!property.value.constant) {
                    argsToWatch.push.apply(argsToWatch, property.value.toWatch);
                }
            }));
            ast.constant = allConstants;
            ast.toWatch = argsToWatch;
            break;
        case AST.prototype.ThisExpression:
            ast.constant = false;
            ast.toWatch = [];
            break;
        case AST.prototype.LocalsExpression:
            ast.constant = false;
            ast.toWatch = [];
            break;
    }
}
/**
 * @param {?} body
 * @return {?}
 */
function getInputs(body) {
    if (body.length !== 1)
        return;
    /** @type {?} */
    var lastExpression = body[0].expression;
    /** @type {?} */
    var candidate = lastExpression.toWatch;
    if (candidate.length !== 1)
        return candidate;
    return candidate[0] !== lastExpression ? candidate : undefined;
}
/**
 * @param {?} ast
 * @return {?}
 */
function isAssignable(ast) {
    return ast.type === AST.prototype.Identifier || ast.type === AST.prototype.MemberExpression;
}
/**
 * @param {?} ast
 * @return {?}
 */
function assignableAST(ast) {
    if (ast.body.length === 1 && isAssignable(ast.body[0].expression)) {
        return {
            type: AST.prototype.AssignmentExpression,
            left: ast.body[0].expression,
            right: { type: AST.prototype.NGValueParameter },
            operator: "="
        };
    }
}
/**
 * @param {?} ast
 * @return {?}
 */
function isLiteral(ast) {
    return (ast.body.length === 0 ||
        (ast.body.length === 1 &&
            (ast.body[0].expression.type === AST.prototype.Literal ||
                ast.body[0].expression.type === AST.prototype.ArrayExpression ||
                ast.body[0].expression.type === AST.prototype.ObjectExpression)));
}
/**
 * @param {?} ast
 * @return {?}
 */
function isConstant(ast) {
    return ast.constant;
}
/**
 * @param {?} astBuilder
 * @param {?} $filter
 * @return {?}
 */
function ASTCompiler(astBuilder, $filter) {
    this.astBuilder = astBuilder;
    this.$filter = $filter;
}
ASTCompiler.prototype = {
    compile: (/**
     * @param {?} expression
     * @return {?}
     */
    function (expression) {
        /** @type {?} */
        var self = this;
        /** @type {?} */
        var ast = this.astBuilder.ast(expression);
        this.state = {
            nextId: 0,
            filters: {},
            fn: { vars: [], body: [], own: {} },
            assign: { vars: [], body: [], own: {} },
            inputs: []
        };
        findConstantAndWatchExpressions(ast, self.$filter);
        /** @type {?} */
        var extra = "";
        /** @type {?} */
        var assignable;
        this.stage = "assign";
        if ((assignable = assignableAST(ast))) {
            this.state.computing = "assign";
            /** @type {?} */
            var result = this.nextId();
            this.recurse(assignable, result);
            this.return_(result);
            extra = "fn.assign=" + this.generateFunction("assign", "s,v,l");
        }
        /** @type {?} */
        var toWatch = getInputs(ast.body);
        self.stage = "inputs";
        if (!!toWatch) {
            toWatch.forEach((/**
             * @param {?} watch
             * @param {?} key
             * @return {?}
             */
            function (watch, key) {
                /** @type {?} */
                var fnKey = "fn" + key;
                self.state[fnKey] = { vars: [], body: [], own: {} };
                self.state.computing = fnKey;
                /** @type {?} */
                var intoId = self.nextId();
                self.recurse(watch, intoId);
                self.return_(intoId);
                self.state.inputs.push(fnKey);
                watch.watchId = key;
            }));
        }
        this.state.computing = "fn";
        this.stage = "main";
        this.recurse(ast);
        /** @type {?} */
        var fnString = 
        // The build and minification steps remove the string "use strict" from the code, but this is done using a regex.
        // This is a workaround for this until we do a better job at only removing the prefix only when we should.
        '"' +
            this.USE +
            " " +
            this.STRICT +
            '";\n' +
            this.filterPrefix() +
            "var fn=" +
            this.generateFunction("fn", "s,l,a,i") +
            extra +
            this.watchFns() +
            "return fn;";
        // eslint-disable-next-line no-new-func
        /** @type {?} */
        var fn = new Function("$filter", "getStringValue", "ifDefined", "plus", fnString)(this.$filter, getStringValue, ifDefined, plusFn);
        this.state = this.stage = undefined;
        fn.ast = ast;
        fn.literal = isLiteral(ast);
        fn.constant = isConstant(ast);
        return fn;
    }),
    USE: "use",
    STRICT: "strict",
    watchFns: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var result = [];
        /** @type {?} */
        var fns = this.state.inputs;
        /** @type {?} */
        var self = this;
        fns.forEach((/**
         * @param {?} name
         * @return {?}
         */
        function (name) {
            result.push("var " + name + "=" + self.generateFunction(name, "s"));
        }));
        if (fns.length) {
            result.push("fn.inputs=[" + fns.join(",") + "];");
        }
        return result.join("");
    }),
    generateFunction: (/**
     * @param {?} name
     * @param {?} params
     * @return {?}
     */
    function (name, params) {
        return ("function(" +
            params +
            "){" +
            this.varsPrefix(name) +
            this.body(name) +
            "};");
    }),
    filterPrefix: (/**
     * @return {?}
     */
    function () {
        /** @type {?} */
        var parts = [];
        /** @type {?} */
        var self = this;
        Object.keys(this.state.filters).forEach((/**
         * @param {?} id
         * @return {?}
         */
        function (id) {
            /** @type {?} */
            let filter = this.state.filters[id];
            parts.push(id + "=$filter(" + self.escape(filter) + ")");
        }));
        if (parts.length)
            return "var " + parts.join(",") + ";";
        return "";
    }),
    varsPrefix: (/**
     * @param {?} section
     * @return {?}
     */
    function (section) {
        return this.state[section].vars.length
            ? "var " + this.state[section].vars.join(",") + ";"
            : "";
    }),
    body: (/**
     * @param {?} section
     * @return {?}
     */
    function (section) {
        return this.state[section].body.join("");
    }),
    recurse: (/**
     * @param {?} ast
     * @param {?} intoId
     * @param {?} nameId
     * @param {?} recursionFn
     * @param {?} create
     * @param {?} skipWatchIdCheck
     * @return {?}
     */
    function (ast, intoId, nameId, recursionFn, create, skipWatchIdCheck) {
        /** @type {?} */
        var left;
        /** @type {?} */
        var right;
        /** @type {?} */
        var self = this;
        /** @type {?} */
        var args;
        /** @type {?} */
        var expression;
        /** @type {?} */
        var computed;
        recursionFn = recursionFn || noop;
        if (!skipWatchIdCheck && isDefined(ast.watchId)) {
            intoId = intoId || this.nextId();
            this.if_("i", this.lazyAssign(intoId, this.computedMember("i", ast.watchId)), this.lazyRecurse(ast, intoId, nameId, recursionFn, create, true));
            return;
        }
        switch (ast.type) {
            case AST.prototype.Program:
                ast.body.forEach((/**
                 * @param {?} expression
                 * @param {?} pos
                 * @return {?}
                 */
                function (expression, pos) {
                    self.recurse(expression.expression, undefined, undefined, (/**
                     * @param {?} expr
                     * @return {?}
                     */
                    function (expr) {
                        right = expr;
                    }));
                    if (pos !== ast.body.length - 1) {
                        self.current().body.push(right, ";");
                    }
                    else {
                        self.return_(right);
                    }
                }));
                break;
            case AST.prototype.Literal:
                expression = this.escape(ast.value);
                this.assign(intoId, expression);
                recursionFn(intoId || expression);
                break;
            case AST.prototype.UnaryExpression:
                this.recurse(ast.argument, undefined, undefined, (/**
                 * @param {?} expr
                 * @return {?}
                 */
                function (expr) {
                    right = expr;
                }));
                expression = ast.operator + "(" + this.ifDefined(right, 0) + ")";
                this.assign(intoId, expression);
                recursionFn(expression);
                break;
            case AST.prototype.BinaryExpression:
                this.recurse(ast.left, undefined, undefined, (/**
                 * @param {?} expr
                 * @return {?}
                 */
                function (expr) {
                    left = expr;
                }));
                this.recurse(ast.right, undefined, undefined, (/**
                 * @param {?} expr
                 * @return {?}
                 */
                function (expr) {
                    right = expr;
                }));
                if (ast.operator === "+") {
                    expression = this.plus(left, right);
                }
                else if (ast.operator === "-") {
                    expression =
                        this.ifDefined(left, 0) + ast.operator + this.ifDefined(right, 0);
                }
                else {
                    expression = "(" + left + ")" + ast.operator + "(" + right + ")";
                }
                this.assign(intoId, expression);
                recursionFn(expression);
                break;
            case AST.prototype.LogicalExpression:
                intoId = intoId || this.nextId();
                self.recurse(ast.left, intoId);
                self.if_(ast.operator === "&&" ? intoId : self.not(intoId), self.lazyRecurse(ast.right, intoId));
                recursionFn(intoId);
                break;
            case AST.prototype.ConditionalExpression:
                intoId = intoId || this.nextId();
                self.recurse(ast.test, intoId);
                self.if_(intoId, self.lazyRecurse(ast.alternate, intoId), self.lazyRecurse(ast.consequent, intoId));
                recursionFn(intoId);
                break;
            case AST.prototype.Identifier:
                intoId = intoId || this.nextId();
                if (nameId) {
                    nameId.context =
                        self.stage === "inputs"
                            ? "s"
                            : this.assign(this.nextId(), this.getHasOwnProperty("l", ast.name) + "?l:s");
                    nameId.computed = false;
                    nameId.name = ast.name;
                }
                self.if_(self.stage === "inputs" ||
                    self.not(self.getHasOwnProperty("l", ast.name)), (/**
                 * @return {?}
                 */
                function () {
                    self.if_(self.stage === "inputs" ||
                        self.and_("s", self.or_(self.isNull(self.nonComputedMember("s", ast.name)), self.hasOwnProperty_("s", ast.name))), (/**
                     * @return {?}
                     */
                    function () {
                        if (create && create !== 1) {
                            self.if_(self.isNull(self.nonComputedMember("s", ast.name)), self.lazyAssign(self.nonComputedMember("s", ast.name), "{}"));
                        }
                        self.assign(intoId, self.nonComputedMember("s", ast.name));
                    }));
                }), intoId &&
                    self.lazyAssign(intoId, self.nonComputedMember("l", ast.name)));
                recursionFn(intoId);
                break;
            case AST.prototype.MemberExpression:
                left = (nameId && (nameId.context = this.nextId())) || this.nextId();
                intoId = intoId || this.nextId();
                self.recurse(ast.object, left, undefined, (/**
                 * @return {?}
                 */
                function () {
                    /** @type {?} */
                    var member = null;
                    if (ast.computed) {
                        right = self.nextId();
                        member = self.computedMember(left, right);
                    }
                    else {
                        member = self.nonComputedMember(left, ast.property.name);
                        right = ast.property.name;
                    }
                    if (ast.computed) {
                        if (ast.property.type === AST.prototype.Literal) {
                            self.recurse(ast.property, right);
                        }
                    }
                    self.if_(self.and_(self.notNull(left), self.or_(self.isNull(member), self.hasOwnProperty_(left, right, ast.computed))), (/**
                     * @return {?}
                     */
                    function () {
                        if (ast.computed) {
                            if (ast.property.type !== AST.prototype.Literal) {
                                self.recurse(ast.property, right);
                            }
                            if (create && create !== 1) {
                                self.if_(self.not(member), self.lazyAssign(member, "{}"));
                            }
                            self.assign(intoId, member);
                            if (nameId) {
                                nameId.computed = true;
                                nameId.name = right;
                            }
                        }
                        else {
                            if (create && create !== 1) {
                                self.if_(self.isNull(member), self.lazyAssign(member, "{}"));
                            }
                            self.assign(intoId, member);
                            if (nameId) {
                                nameId.computed = false;
                                nameId.name = ast.property.name;
                            }
                        }
                    }), (/**
                     * @return {?}
                     */
                    function () {
                        self.assign(intoId, "undefined");
                    }));
                    recursionFn(intoId);
                }), !!create);
                break;
            case AST.prototype.CallExpression:
                intoId = intoId || this.nextId();
                if (ast.filter) {
                    right = self.filter(ast.callee.name);
                    args = [];
                    ast.arguments.forEach((/**
                     * @param {?} expr
                     * @return {?}
                     */
                    function (expr) {
                        /** @type {?} */
                        var argument = self.nextId();
                        self.recurse(expr, argument);
                        args.push(argument);
                    }));
                    expression = right + ".call(" + right + "," + args.join(",") + ")";
                    self.assign(intoId, expression);
                    recursionFn(intoId);
                }
                else {
                    right = self.nextId();
                    left = {};
                    args = [];
                    self.recurse(ast.callee, right, left, (/**
                     * @return {?}
                     */
                    function () {
                        self.if_(self.notNull(right), (/**
                         * @return {?}
                         */
                        function () {
                            ast.arguments.forEach((/**
                             * @param {?} expr
                             * @return {?}
                             */
                            function (expr) {
                                self.recurse(expr, ast.constant ? undefined : self.nextId(), undefined, (/**
                                 * @param {?} argument
                                 * @return {?}
                                 */
                                function (argument) {
                                    args.push(argument);
                                }));
                            }));
                            if (left.name) {
                                /** @type {?} */
                                var x = self.member(left.context, left.name, left.computed);
                                expression =
                                    x + ".call(" + [left.context].concat(args).join(",") + ")";
                            }
                            else {
                                expression = right + "(" + args.join(",") + ")";
                            }
                            self.assign(intoId, expression);
                        }), (/**
                         * @return {?}
                         */
                        function () {
                            self.assign(intoId, "undefined");
                        }));
                        recursionFn(intoId);
                    }));
                }
                break;
            case AST.prototype.AssignmentExpression:
                right = this.nextId();
                left = {};
                this.recurse(ast.left, undefined, left, (/**
                 * @return {?}
                 */
                function () {
                    self.if_(self.and_(self.notNull(left.context), self.or_(self.hasOwnProperty_(left.context, left.name), self.isNull(self.member(left.context, left.name, left.computed)))), (/**
                     * @return {?}
                     */
                    function () {
                        self.recurse(ast.right, right);
                        expression =
                            self.member(left.context, left.name, left.computed) +
                                ast.operator +
                                right;
                        self.assign(intoId, expression);
                        recursionFn(intoId || expression);
                    }));
                }), 1);
                break;
            case AST.prototype.ArrayExpression:
                args = [];
                ast.elements.forEach((/**
                 * @param {?} expr
                 * @return {?}
                 */
                function (expr) {
                    self.recurse(expr, ast.constant ? undefined : self.nextId(), undefined, (/**
                     * @param {?} argument
                     * @return {?}
                     */
                    function (argument) {
                        args.push(argument);
                    }));
                }));
                expression = "[" + args.join(",") + "]";
                this.assign(intoId, expression);
                recursionFn(intoId || expression);
                break;
            case AST.prototype.ObjectExpression:
                args = [];
                computed = false;
                ast.properties.forEach((/**
                 * @param {?} property
                 * @return {?}
                 */
                function (property) {
                    if (property.computed) {
                        computed = true;
                    }
                }));
                if (computed) {
                    intoId = intoId || this.nextId();
                    this.assign(intoId, "{}");
                    ast.properties.forEach((/**
                     * @param {?} property
                     * @return {?}
                     */
                    function (property) {
                        if (property.computed) {
                            left = self.nextId();
                            self.recurse(property.key, left);
                        }
                        else {
                            left =
                                property.key.type === AST.prototype.Identifier
                                    ? property.key.name
                                    : "" + property.key.value;
                        }
                        right = self.nextId();
                        self.recurse(property.value, right);
                        self.assign(self.member(intoId, left, property.computed), right);
                    }));
                }
                else {
                    ast.properties.forEach((/**
                     * @param {?} property
                     * @return {?}
                     */
                    function (property) {
                        self.recurse(property.value, ast.constant ? undefined : self.nextId(), undefined, (/**
                         * @param {?} expr
                         * @return {?}
                         */
                        function (expr) {
                            args.push(self.escape(property.key.type === AST.prototype.Identifier
                                ? property.key.name
                                : "" + property.key.value) +
                                ":" +
                                expr);
                        }));
                    }));
                    expression = "{" + args.join(",") + "}";
                    this.assign(intoId, expression);
                }
                recursionFn(intoId || expression);
                break;
            case AST.prototype.ThisExpression:
                this.assign(intoId, "s");
                recursionFn(intoId || "s");
                break;
            case AST.prototype.LocalsExpression:
                this.assign(intoId, "l");
                recursionFn(intoId || "l");
                break;
            case AST.prototype.NGValueParameter:
                this.assign(intoId, "v");
                recursionFn(intoId || "v");
                break;
        }
    }),
    getHasOwnProperty: (/**
     * @param {?} element
     * @param {?} property
     * @return {?}
     */
    function (element, property) {
        /** @type {?} */
        var key = element + "." + property;
        /** @type {?} */
        var own = this.current().own;
        if (!own.hasOwnProperty(key)) {
            own[key] = this.nextId(false, element + "&&(" + this.escape(property) + " in " + element + ")");
        }
        return own[key];
    }),
    assign: (/**
     * @param {?} id
     * @param {?} value
     * @return {?}
     */
    function (id, value) {
        if (!id)
            return;
        this.current().body.push(id, "=", value, ";");
        return id;
    }),
    filter: (/**
     * @param {?} filterName
     * @return {?}
     */
    function (filterName) {
        if (!this.state.filters.hasOwnProperty(filterName)) {
            this.state.filters[filterName] = this.nextId(true);
        }
        return this.state.filters[filterName];
    }),
    ifDefined: (/**
     * @param {?} id
     * @param {?} defaultValue
     * @return {?}
     */
    function (id, defaultValue) {
        return "ifDefined(" + id + "," + this.escape(defaultValue) + ")";
    }),
    plus: (/**
     * @param {?} left
     * @param {?} right
     * @return {?}
     */
    function (left, right) {
        return "plus(" + left + "," + right + ")";
    }),
    return_: (/**
     * @param {?} id
     * @return {?}
     */
    function (id) {
        this.current().body.push("return ", id, ";");
    }),
    if_: (/**
     * @param {?} test
     * @param {?} alternate
     * @param {?} consequent
     * @return {?}
     */
    function (test, alternate, consequent) {
        if (test === true) {
            alternate();
        }
        else {
            /** @type {?} */
            var body = this.current().body;
            body.push("if(", test, "){");
            alternate();
            body.push("}");
            if (consequent) {
                body.push("else{");
                consequent();
                body.push("}");
            }
        }
    }),
    or_: (/**
     * @param {?} expr1
     * @param {?} expr2
     * @return {?}
     */
    function (expr1, expr2) {
        return "(" + expr1 + ") || (" + expr2 + ")";
    }),
    hasOwnProperty_: (/**
     * @param {?} obj
     * @param {?} prop
     * @param {?} computed
     * @return {?}
     */
    function (obj, prop, computed) {
        if (computed) {
            return "(Object.prototype.hasOwnProperty.call(" + obj + "," + prop + "))";
        }
        else {
            return ("(Object.prototype.hasOwnProperty.call(" + obj + ",'" + prop + "'))");
        }
    }),
    and_: (/**
     * @param {?} expr1
     * @param {?} expr2
     * @return {?}
     */
    function (expr1, expr2) {
        return "(" + expr1 + ") && (" + expr2 + ")";
    }),
    not: (/**
     * @param {?} expression
     * @return {?}
     */
    function (expression) {
        return "!(" + expression + ")";
    }),
    isNull: (/**
     * @param {?} expression
     * @return {?}
     */
    function (expression) {
        return expression + "==null";
    }),
    notNull: (/**
     * @param {?} expression
     * @return {?}
     */
    function (expression) {
        return expression + "!=null";
    }),
    nonComputedMember: (/**
     * @param {?} left
     * @param {?} right
     * @return {?}
     */
    function (left, right) {
        /** @type {?} */
        var SAFE_IDENTIFIER = /^[$_a-zA-Z][$_a-zA-Z0-9]*$/;
        /** @type {?} */
        var UNSAFE_CHARACTERS = /[^$_a-zA-Z0-9]/g;
        /** @type {?} */
        var expr = "";
        if (SAFE_IDENTIFIER.test(right)) {
            expr = left + "." + right;
        }
        else {
            right = right.replace(UNSAFE_CHARACTERS, this.stringEscapeFn);
            expr = left + '["' + right + '"]';
        }
        return expr;
    }),
    computedMember: (/**
     * @param {?} left
     * @param {?} right
     * @return {?}
     */
    function (left, right) {
        return left + "[" + right + "]";
    }),
    member: (/**
     * @param {?} left
     * @param {?} right
     * @param {?} computed
     * @return {?}
     */
    function (left, right, computed) {
        if (computed)
            return this.computedMember(left, right);
        return this.nonComputedMember(left, right);
    }),
    getStringValue: (/**
     * @param {?} item
     * @return {?}
     */
    function (item) {
        this.assign(item, "getStringValue(" + item + ")");
    }),
    lazyRecurse: (/**
     * @param {?} ast
     * @param {?} intoId
     * @param {?} nameId
     * @param {?} recursionFn
     * @param {?} create
     * @param {?} skipWatchIdCheck
     * @return {?}
     */
    function (ast, intoId, nameId, recursionFn, create, skipWatchIdCheck) {
        /** @type {?} */
        var self = this;
        return (/**
         * @return {?}
         */
        function () {
            self.recurse(ast, intoId, nameId, recursionFn, create, skipWatchIdCheck);
        });
    }),
    lazyAssign: (/**
     * @param {?} id
     * @param {?} value
     * @return {?}
     */
    function (id, value) {
        /** @type {?} */
        var self = this;
        return (/**
         * @return {?}
         */
        function () {
            self.assign(id, value);
        });
    }),
    stringEscapeRegex: /[^ a-zA-Z0-9]/g,
    stringEscapeFn: (/**
     * @param {?} c
     * @return {?}
     */
    function (c) {
        return "\\u" + ("0000" + c.charCodeAt(0).toString(16)).slice(-4);
    }),
    escape: (/**
     * @param {?} value
     * @return {?}
     */
    function (value) {
        if (isString(value))
            return ("'" + value.replace(this.stringEscapeRegex, this.stringEscapeFn) + "'");
        if (isNumber(value))
            return value.toString();
        if (value === true)
            return "true";
        if (value === false)
            return "false";
        if (value === null)
            return "null";
        if (typeof value === "undefined")
            return "undefined";
        throw this.throwError("esc", "IMPOSSIBLE");
    }),
    nextId: (/**
     * @param {?} skip
     * @param {?} init
     * @return {?}
     */
    function (skip, init) {
        /** @type {?} */
        var id = "v" + this.state.nextId++;
        if (!skip) {
            this.current().vars.push(id + (init ? "=" + init : ""));
        }
        return id;
    }),
    current: (/**
     * @return {?}
     */
    function () {
        return this.state[this.state.computing];
    })
};
/**
 * @param {?} astBuilder
 * @param {?} $filter
 * @return {?}
 */
function ASTInterpreter(astBuilder, $filter) {
    this.astBuilder = astBuilder;
    this.$filter = $filter;
}
ASTInterpreter.prototype = {
    compile: (/**
     * @param {?} expression
     * @return {?}
     */
    function (expression) {
        /** @type {?} */
        var self = this;
        /** @type {?} */
        var ast = this.astBuilder.ast(expression);
        findConstantAndWatchExpressions(ast, self.$filter);
        /** @type {?} */
        var assignable;
        /** @type {?} */
        var assign;
        if ((assignable = assignableAST(ast))) {
            assign = this.recurse(assignable);
        }
        /** @type {?} */
        var toWatch = getInputs(ast.body);
        /** @type {?} */
        var inputs;
        if (toWatch) {
            inputs = [];
            toWatch.forEach((/**
             * @param {?} watch
             * @param {?} key
             * @return {?}
             */
            function (watch, key) {
                /** @type {?} */
                var input = self.recurse(watch);
                watch.input = input;
                inputs.push(input);
                watch.watchId = key;
            }));
        }
        /** @type {?} */
        var expressions = [];
        ast.body.forEach((/**
         * @param {?} expression
         * @return {?}
         */
        function (expression) {
            expressions.push(self.recurse(expression.expression));
        }));
        /** @type {?} */
        var fn = ast.body.length === 0
            ? noop
            : ast.body.length === 1
                ? expressions[0]
                : (/**
                 * @param {?} scope
                 * @param {?} locals
                 * @return {?}
                 */
                function (scope, locals) {
                    /** @type {?} */
                    var lastValue;
                    expressions.forEach((/**
                     * @param {?} exp
                     * @return {?}
                     */
                    function (exp) {
                        lastValue = exp(scope, locals);
                    }));
                    return lastValue;
                });
        if (assign) {
            fn.assign = (/**
             * @param {?} scope
             * @param {?} value
             * @param {?} locals
             * @return {?}
             */
            function (scope, value, locals) {
                return assign(scope, locals, value);
            });
        }
        if (inputs) {
            fn.inputs = inputs;
        }
        fn.ast = ast;
        fn.literal = isLiteral(ast);
        fn.constant = isConstant(ast);
        return fn;
    }),
    recurse: (/**
     * @param {?} ast
     * @param {?} context
     * @param {?} create
     * @return {?}
     */
    function (ast, context, create) {
        /** @type {?} */
        var left;
        /** @type {?} */
        var right;
        /** @type {?} */
        var self = this;
        /** @type {?} */
        var args;
        if (ast.input) {
            return this.inputs(ast.input, ast.watchId);
        }
        switch (ast.type) {
            case AST.prototype.Literal:
                return this.value(ast.value, context);
            case AST.prototype.UnaryExpression:
                right = this.recurse(ast.argument);
                return this["unary" + ast.operator](right, context);
            case AST.prototype.BinaryExpression:
                left = this.recurse(ast.left);
                right = this.recurse(ast.right);
                return this["binary" + ast.operator](left, right, context);
            case AST.prototype.LogicalExpression:
                left = this.recurse(ast.left);
                right = this.recurse(ast.right);
                return this["binary" + ast.operator](left, right, context);
            case AST.prototype.ConditionalExpression:
                return this["ternary?:"](this.recurse(ast.test), this.recurse(ast.alternate), this.recurse(ast.consequent), context);
            case AST.prototype.Identifier:
                return self.identifier(ast.name, context, create);
            case AST.prototype.MemberExpression:
                left = this.recurse(ast.object, false, !!create);
                if (!ast.computed) {
                    right = ast.property.name;
                }
                if (ast.computed)
                    right = this.recurse(ast.property);
                return ast.computed
                    ? this.computedMember(left, right, context, create)
                    : this.nonComputedMember(left, right, context, create);
            case AST.prototype.CallExpression:
                args = [];
                ast.arguments.forEach((/**
                 * @param {?} expr
                 * @return {?}
                 */
                function (expr) {
                    args.push(self.recurse(expr));
                }));
                if (ast.filter)
                    right = this.$filter(ast.callee.name);
                if (!ast.filter)
                    right = this.recurse(ast.callee, true);
                return ast.filter
                    ? (/**
                     * @param {?} scope
                     * @param {?} locals
                     * @param {?} assign
                     * @param {?} inputs
                     * @return {?}
                     */
                    function (scope, locals, assign, inputs) {
                        /** @type {?} */
                        var values = [];
                        for (var i = 0; i < args.length; ++i) {
                            values.push(args[i](scope, locals, assign, inputs));
                        }
                        /** @type {?} */
                        var value = right.apply(undefined, values, inputs);
                        return context
                            ? { context: undefined, name: undefined, value: value }
                            : value;
                    })
                    : (/**
                     * @param {?} scope
                     * @param {?} locals
                     * @param {?} assign
                     * @param {?} inputs
                     * @return {?}
                     */
                    function (scope, locals, assign, inputs) {
                        /** @type {?} */
                        var rhs = right(scope, locals, assign, inputs);
                        /** @type {?} */
                        var value;
                        if (rhs.value != null) {
                            /** @type {?} */
                            var values = [];
                            for (var i = 0; i < args.length; ++i) {
                                values.push(args[i](scope, locals, assign, inputs));
                            }
                            value = rhs.value.apply(rhs.context, values);
                        }
                        return context ? { value: value } : value;
                    });
            case AST.prototype.AssignmentExpression:
                left = this.recurse(ast.left, true, 1);
                right = this.recurse(ast.right);
                return (/**
                 * @param {?} scope
                 * @param {?} locals
                 * @param {?} assign
                 * @param {?} inputs
                 * @return {?}
                 */
                function (scope, locals, assign, inputs) {
                    /** @type {?} */
                    var lhs = left(scope, locals, assign, inputs);
                    /** @type {?} */
                    var rhs = right(scope, locals, assign, inputs);
                    lhs.context[lhs.name] = rhs;
                    return context ? { value: rhs } : rhs;
                });
            case AST.prototype.ArrayExpression:
                args = [];
                ast.elements.forEach((/**
                 * @param {?} expr
                 * @return {?}
                 */
                function (expr) {
                    args.push(self.recurse(expr));
                }));
                return (/**
                 * @param {?} scope
                 * @param {?} locals
                 * @param {?} assign
                 * @param {?} inputs
                 * @return {?}
                 */
                function (scope, locals, assign, inputs) {
                    /** @type {?} */
                    var value = [];
                    for (var i = 0; i < args.length; ++i) {
                        value.push(args[i](scope, locals, assign, inputs));
                    }
                    return context ? { value: value } : value;
                });
            case AST.prototype.ObjectExpression:
                args = [];
                ast.properties.forEach((/**
                 * @param {?} property
                 * @return {?}
                 */
                function (property) {
                    if (property.computed) {
                        args.push({
                            key: self.recurse(property.key),
                            computed: true,
                            value: self.recurse(property.value)
                        });
                    }
                    else {
                        args.push({
                            key: property.key.type === AST.prototype.Identifier
                                ? property.key.name
                                : "" + property.key.value,
                            computed: false,
                            value: self.recurse(property.value)
                        });
                    }
                }));
                return (/**
                 * @param {?} scope
                 * @param {?} locals
                 * @param {?} assign
                 * @param {?} inputs
                 * @return {?}
                 */
                function (scope, locals, assign, inputs) {
                    /** @type {?} */
                    var value = {};
                    for (var i = 0; i < args.length; ++i) {
                        if (args[i].computed) {
                            value[args[i].key(scope, locals, assign, inputs)] = args[i].value(scope, locals, assign, inputs);
                        }
                        else {
                            value[args[i].key] = args[i].value(scope, locals, assign, inputs);
                        }
                    }
                    return context ? { value: value } : value;
                });
            case AST.prototype.ThisExpression:
                return (/**
                 * @param {?} scope
                 * @return {?}
                 */
                function (scope) {
                    return context ? { value: scope } : scope;
                });
            case AST.prototype.LocalsExpression:
                return (/**
                 * @param {?} scope
                 * @param {?} locals
                 * @return {?}
                 */
                function (scope, locals) {
                    return context ? { value: locals } : locals;
                });
            case AST.prototype.NGValueParameter:
                return (/**
                 * @param {?} scope
                 * @param {?} locals
                 * @param {?} assign
                 * @return {?}
                 */
                function (scope, locals, assign) {
                    return context ? { value: assign } : assign;
                });
        }
    }),
    "unary+": (/**
     * @param {?} argument
     * @param {?} context
     * @return {?}
     */
    function (argument, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = argument(scope, locals, assign, inputs);
            if (isDefined(arg)) {
                arg = +arg;
            }
            else {
                arg = 0;
            }
            return context ? { value: arg } : arg;
        });
    }),
    "unary-": (/**
     * @param {?} argument
     * @param {?} context
     * @return {?}
     */
    function (argument, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = argument(scope, locals, assign, inputs);
            if (isDefined(arg)) {
                arg = -arg;
            }
            else {
                arg = -0;
            }
            return context ? { value: arg } : arg;
        });
    }),
    "unary!": (/**
     * @param {?} argument
     * @param {?} context
     * @return {?}
     */
    function (argument, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = !argument(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary+": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var lhs = left(scope, locals, assign, inputs);
            /** @type {?} */
            var rhs = right(scope, locals, assign, inputs);
            /** @type {?} */
            var arg = plusFn(lhs, rhs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary-": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var lhs = left(scope, locals, assign, inputs);
            /** @type {?} */
            var rhs = right(scope, locals, assign, inputs);
            /** @type {?} */
            var arg = (isDefined(lhs) ? lhs : 0) - (isDefined(rhs) ? rhs : 0);
            return context ? { value: arg } : arg;
        });
    }),
    "binary*": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) *
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary/": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) /
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary%": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) %
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary===": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) ===
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary!==": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) !==
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary==": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            // eslint-disable-next-line eqeqeq
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) ==
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary!=": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            // eslint-disable-next-line eqeqeq
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) !=
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary<": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) <
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary>": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) >
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary<=": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) <=
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary>=": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) >=
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary&&": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) &&
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "binary||": (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @return {?}
     */
    function (left, right, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = left(scope, locals, assign, inputs) ||
                right(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    "ternary?:": (/**
     * @param {?} test
     * @param {?} alternate
     * @param {?} consequent
     * @param {?} context
     * @return {?}
     */
    function (test, alternate, consequent, context) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var arg = test(scope, locals, assign, inputs)
                ? alternate(scope, locals, assign, inputs)
                : consequent(scope, locals, assign, inputs);
            return context ? { value: arg } : arg;
        });
    }),
    value: (/**
     * @param {?} value
     * @param {?} context
     * @return {?}
     */
    function (value, context) {
        return (/**
         * @return {?}
         */
        function () {
            return context
                ? { context: undefined, name: undefined, value: value }
                : value;
        });
    }),
    identifier: (/**
     * @param {?} name
     * @param {?} context
     * @param {?} create
     * @return {?}
     */
    function (name, context, create) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var base = locals && name in locals ? locals : scope;
            if (create && create !== 1 && base && base[name] == null) {
                base[name] = {};
            }
            /** @type {?} */
            var value = base ? base[name] : undefined;
            if (context) {
                return { context: base, name: name, value: value };
            }
            else {
                return value;
            }
        });
    }),
    computedMember: (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @param {?} create
     * @return {?}
     */
    function (left, right, context, create) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var lhs = left(scope, locals, assign, inputs);
            /** @type {?} */
            var rhs;
            /** @type {?} */
            var value;
            if (lhs != null) {
                rhs = right(scope, locals, assign, inputs);
                rhs = getStringValue(rhs);
                if (create && create !== 1) {
                    if (lhs && !lhs[rhs]) {
                        lhs[rhs] = {};
                    }
                }
                if (Object.prototype.hasOwnProperty.call(lhs, rhs)) {
                    value = lhs[rhs];
                }
            }
            if (context) {
                return { context: lhs, name: rhs, value: value };
            }
            else {
                return value;
            }
        });
    }),
    nonComputedMember: (/**
     * @param {?} left
     * @param {?} right
     * @param {?} context
     * @param {?} create
     * @return {?}
     */
    function (left, right, context, create) {
        return (/**
         * @param {?} scope
         * @param {?} locals
         * @param {?} assign
         * @param {?} inputs
         * @return {?}
         */
        function (scope, locals, assign, inputs) {
            /** @type {?} */
            var lhs = left(scope, locals, assign, inputs);
            if (create && create !== 1) {
                if (lhs && lhs[right] == null) {
                    lhs[right] = {};
                }
            }
            /** @type {?} */
            var value = undefined;
            if (lhs != null && Object.prototype.hasOwnProperty.call(lhs, right)) {
                value = lhs[right];
            }
            if (context) {
                return { context: lhs, name: right, value: value };
            }
            else {
                return value;
            }
        });
    }),
    inputs: (/**
     * @param {?} input
     * @param {?} watchId
     * @return {?}
     */
    function (input, watchId) {
        return (/**
         * @param {?} scope
         * @param {?} value
         * @param {?} locals
         * @param {?} inputs
         * @return {?}
         */
        function (scope, value, locals, inputs) {
            if (inputs)
                return inputs[watchId];
            return input(scope, value, locals);
        });
    })
};
/**
 * constructor
 * @param {?} lexer
 * @param {?} $filter
 * @param {?} options
 * @return {?}
 */
function Parser(lexer, $filter, options) {
    this.lexer = lexer;
    this.$filter = $filter;
    this.options = options;
    this.ast = new AST(lexer, options);
    this.astCompiler = options.csp
        ? new ASTInterpreter(this.ast, $filter)
        : new ASTCompiler(this.ast, $filter);
}
;
Parser.prototype = {
    constructor: Parser,
    parse: (/**
     * @param {?} text
     * @return {?}
     */
    function (text) {
        return this.astCompiler.compile(text);
    })
};
/**
 * @param {?} value
 * @return {?}
 */
function getValueOf(value) {
    return isFunction(value.valueOf)
        ? value.valueOf()
        : objectValueOf.call(value);
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/expression_parser/parse_main.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
"use strict";
/** @type {?} */
var filters = {};
/**
 * Compiles src and returns a function that executes src on a target object.
 * The compiled function is cached under compile.cache[src] to speed up further calls.
 *
 * param {string} src
 * returns {function}
 * @param {?} src
 * @param {?=} lexerOptions
 * @return {?}
 */
function compile(src, lexerOptions) {
    lexerOptions = lexerOptions || {};
    /** @type {?} */
    var cached;
    if (typeof src !== "string") {
        throw new TypeError("src must be a string, instead saw '" + typeof src + "'");
    }
    /** @type {?} */
    var parserOptions = {
        csp: false,
        // noUnsafeEval,
        expensiveChecks: true,
        literals: {
            // defined at: function $ParseProvider() {
            true: true,
            false: false,
            null: null,
            /*eslint no-undefined: 0*/
            undefined: undefined
            /* eslint: no-undefined: 1  */
        },
    };
    /** @type {?} */
    var lexer = new Lexer(lexerOptions);
    /** @type {?} */
    var parser = new Parser(lexer, (/**
     * @param {?} name
     * @return {?}
     */
    function getFilter(name) {
        return filters[name];
    }), parserOptions);
    if (!compile.cache) {
        return parser.parse(src);
    }
    cached = compile.cache[src];
    if (!cached) {
        cached = compile.cache[src] = parser.parse(src);
    }
    return cached;
}
/**
 * A cache containing all compiled functions. The src is used as key.
 * Set this on false to disable the cache.
 *
 * @type {object}
 */
compile.cache = Object.create(null);

/**
 * @fileoverview added by tsickle
 * Generated from: lib/service/ngx-validation-runner.service.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxValidationRunnerService {
    /**
     * @param {?} ngZone
     */
    constructor(ngZone) {
        this.ngZone = ngZone;
        this.policies = {};
        this.validationListeners = {};
        this.requiredListeners = {};
        this._requiredCheckQueue = [];
        this._executeValidation = (/**
         * @param {?} policy
         * @param {?} model
         * @param {?} validationModelResult
         * @return {?}
         */
        (policy, model, validationModelResult) => __awaiter(this, void 0, void 0, function* () {
            /** @type {?} */
            let dependencyOutput = true;
            if (policy.dependency) {
                if (typeof (policy.dependency) === 'function') {
                    dependencyOutput = yield policy.dependency.call(this, model);
                }
                else {
                    /** @type {?} */
                    const comFn = compile(policy.dependency);
                    /** @type {?} */
                    const dependentValueEvaluated = comFn(model);
                    dependencyOutput = dependentValueEvaluated;
                }
            }
            // if no dependency
            if (dependencyOutput === true) {
                /** @type {?} */
                const comFn = compile(policy.propertyName);
                /** @type {?} */
                const propValueEvaluated = comFn(model);
                if (policy.validationFns.length > 0) {
                    // `${policy.propertyName.split('.').join(' ')} is required`
                    yield policy.validationFns[0].call(this, propValueEvaluated, policy, model);
                }
                if (!!policy.errorMsg) {
                    // add the first error
                    validationModelResult.push(this.createError(policy.propertyName, policy.errorMsg));
                    return;
                }
                else {
                    // loop through rest of the validation functions if error not found
                    for (let cur = 1; cur < policy.validationFns.length; cur++) {
                        yield policy.validationFns[cur].call(this, propValueEvaluated, policy, model);
                        if (!!policy.errorMsg) {
                            validationModelResult.push(this.createError(policy.propertyName, policy.errorMsg));
                        }
                    }
                }
            }
        }));
        this._executeRequiredCheckOnModel = (/**
         * @param {?} policy
         * @param {?} model
         * @param {?} requiredModelResult
         * @return {?}
         */
        (policy, model, requiredModelResult) => {
            /** @type {?} */
            let dependencyOutput = true;
            if (policy.dependency) {
                if (typeof (policy.dependency) === 'function') {
                    dependencyOutput = policy.dependency.call(this, model);
                }
                else {
                    /** @type {?} */
                    const comFn = compile(policy.dependency);
                    /** @type {?} */
                    const dependentValueEvaluated = comFn(model);
                    dependencyOutput = dependentValueEvaluated;
                }
            }
            // if no dependency
            if (dependencyOutput === true) {
                // means required
                requiredModelResult.push(this.createError(policy.propertyName, 'required'));
            }
        });
        this.ngZone.runOutsideAngular((/**
         * @return {?}
         */
        () => {
            setInterval((/**
             * @return {?}
             */
            () => {
                if (this._requiredCheckQueue.length > 0) {
                    /** @type {?} */
                    const callArgs = this._requiredCheckQueue[this._requiredCheckQueue.length - 1];
                    this._requiredCheckQueue = [];
                    this._runRequiredCheck.call(this, ...callArgs);
                }
            }), 1000);
        }));
    }
    /**
     * @param {?} policyName
     * @param {?} policy
     * @return {?}
     */
    register(policyName, policy) {
        /** @type {?} */
        const existingPolicy = this.getPolicy(policyName);
        if (!!existingPolicy && existingPolicy.length > 0) {
            console.warn(`${policyName} is already registered`);
            return existingPolicy;
        }
        /** @type {?} */
        const validators = policy.addValidations(new NgxValidatorHelper());
        this.policies[policyName] = validators;
        // for validation emit from svc to component validator
        this.validationListeners[policyName] = new BehaviorSubject((/** @type {?} */ ({})));
        // for required emit from svc to component validator
        this.requiredListeners[policyName] = new Subject();
        return [...this.policies[policyName]];
    }
    /**
     * @param {?} policyName
     * @return {?}
     */
    getPolicy(policyName) {
        if (this.policies[policyName]) {
            return [...this.policies[policyName]];
        }
        else {
            return [];
        }
    }
    /**
     * @param {?} policyName
     * @return {?}
     */
    getValidationListener(policyName) {
        return this.validationListeners[policyName].asObservable().pipe(share());
    }
    /**
     * @param {?} policyName
     * @return {?}
     */
    getRequiredCheckListener(policyName) {
        return this.requiredListeners[policyName].asObservable().pipe(share());
    }
    /**
     * @param {?} policyName
     * @param {?} model
     * @param {?=} propertyName
     * @return {?}
     */
    validate(policyName, model, propertyName) {
        if (!policyName) {
            throw new Error('policyName is required and should be passed');
        }
        /** @type {?} */
        const policies = this.getPolicy(policyName);
        /** @type {?} */
        const validationSubject = this.validationListeners[policyName];
        /** @type {?} */
        let validationModelResult = [];
        /** @type {?} */
        let validatorsExcutedCount = 0;
        /** @type {?} */
        const responseSubject = new Subject();
        if (propertyName) {
            // read previous result from behavior subject
            validationModelResult = !!validationSubject.value.errors ? validationSubject.value.errors : [];
            validationModelResult = validationModelResult.filter((/**
             * @param {?} item
             * @return {?}
             */
            item => item.propertyName !== propertyName));
            /** @type {?} */
            let policy = policies.find((/**
             * @param {?} item
             * @return {?}
             */
            item => item.propertyName === propertyName));
            if (policy) {
                // using IIFE
                ((/**
                 * @return {?}
                 */
                () => __awaiter(this, void 0, void 0, function* () {
                    this._executeValidation(policy, model, validationModelResult).then((/**
                     * @return {?}
                     */
                    () => {
                        ++validatorsExcutedCount;
                        if (validatorsExcutedCount === 1) {
                            /** @type {?} */
                            const response = (/** @type {?} */ ({
                                type: 'validation',
                                errors: validationModelResult
                            }));
                            validationSubject.next(response);
                            responseSubject.next(validationModelResult);
                        }
                    }));
                })))();
            }
        }
        else {
            Object.keys(policies).forEach((/**
             * @param {?} key
             * @return {?}
             */
            (key) => {
                /** @type {?} */
                let policy = policies[key];
                if (policy) {
                    // using IIFE
                    ((/**
                     * @return {?}
                     */
                    () => __awaiter(this, void 0, void 0, function* () {
                        this._executeValidation(policy, model, validationModelResult).then((/**
                         * @return {?}
                         */
                        () => {
                            ++validatorsExcutedCount;
                            if (validatorsExcutedCount === policies.length) {
                                /** @type {?} */
                                const response = (/** @type {?} */ ({
                                    type: 'validation',
                                    errors: validationModelResult
                                }));
                                validationSubject.next(response);
                                responseSubject.next(validationModelResult);
                            }
                        }));
                    })))();
                }
            }), this);
        }
        return responseSubject.asObservable();
    }
    /**
     * @param {?} policyName
     * @param {?} model
     * @return {?}
     */
    markForRequiredCheck(policyName, model) {
        this._requiredCheckQueue.push([policyName, model]);
    }
    /**
     * @private
     * @param {?} policyName
     * @param {?} model
     * @return {?}
     */
    _runRequiredCheck(policyName, model) {
        if (!policyName) {
            throw new Error('policyName is required and should be passed');
        }
        /** @type {?} */
        const policies = this.getPolicy(policyName);
        /** @type {?} */
        const requiredCheckSubject = this.requiredListeners[policyName];
        // no need to read from previous value
        /** @type {?} */
        let requiredModelResult = [];
        Object.keys(policies).forEach((/**
         * @param {?} key
         * @return {?}
         */
        (key) => {
            /** @type {?} */
            let policy = policies[key];
            if (policy) {
                this._executeRequiredCheckOnModel(policy, model, requiredModelResult);
            }
        }), this);
        /** @type {?} */
        const response = (/** @type {?} */ ({
            type: 'required',
            errors: requiredModelResult
        }));
        requiredCheckSubject.next(response);
        return (of(requiredModelResult));
    }
    /**
     * @param {?} propertyName
     * @param {?} errorMessage
     * @return {?}
     */
    createError(propertyName, errorMessage) {
        return (/** @type {?} */ ({
            propertyName,
            errorMessage
        }));
    }
}
NgxValidationRunnerService.decorators = [
    { type: Injectable }
];
/** @nocollapse */
NgxValidationRunnerService.ctorParameters = () => [
    { type: NgZone }
];
if (false) {
    /**
     * @type {?}
     * @private
     */
    NgxValidationRunnerService.prototype.policies;
    /**
     * @type {?}
     * @private
     */
    NgxValidationRunnerService.prototype.validationListeners;
    /**
     * @type {?}
     * @private
     */
    NgxValidationRunnerService.prototype.requiredListeners;
    /**
     * @type {?}
     * @private
     */
    NgxValidationRunnerService.prototype._requiredCheckQueue;
    /**
     * @type {?}
     * @private
     */
    NgxValidationRunnerService.prototype._executeValidation;
    /**
     * @type {?}
     * @private
     */
    NgxValidationRunnerService.prototype._executeRequiredCheckOnModel;
    /**
     * @type {?}
     * @private
     */
    NgxValidationRunnerService.prototype.ngZone;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/util/window-ref.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const WINDOW = new InjectionToken('WindowToken', typeof window !== 'undefined' && window.document ? { providedIn: 'root', factory: (/**
     * @return {?}
     */
    () => window) } : undefined);

/**
 * @fileoverview added by tsickle
 * Generated from: lib/ngx-material-validator.directive.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// TODO : abstract the validation logic in a base class and use it in a child class for
// any type of UI, e.g. material, bootstrap etc.
// usage    
//   ngxMatValidate [validateProperty]="person.age" [model]="person" [policy]="PERSON_POLICY_NAME" [(ngModel)]="person.age"
// try follow https://github.com/rsaenen/ngx-custom-validators/blob/master/src/app/less-than/directive.ts
class NgxMatValidatorDirective {
    /**
     * @param {?} validationService
     * @param {?} elementRef
     * @param {?} ngZone
     * @param {?} renderer2
     * @param {?} _window
     */
    constructor(validationService, elementRef, ngZone, renderer2, /* private ngModel: NgModel,*/ _window) {
        this.validationService = validationService;
        this.elementRef = elementRef;
        this.ngZone = ngZone;
        this.renderer2 = renderer2;
        this._window = _window;
        this.validateProperty = {};
        this.parentNodeName = '';
        this.isInput = false;
        this.isCheckbox = false;
        this.isRadioButton = false;
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        /**
         * add message event listener
         * message shall come from component validator directive
         */
        // https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage
        this.ngZone.runOutsideAngular((/**
         * @return {?}
         */
        () => {
            this._window.addEventListener("message", (/**
             * @param {?} event
             * @return {?}
             */
            (event) => {
                if (event.data.type === 'validation' && event.data.propertyName === this.validateProperty) {
                    /**
                     * should I skip validation if control is disabled ??
                     */
                    this.addErrorToMatControl(event.data.operation === 'added' ? event.data.errorMessage : null);
                }
            }), false);
        }));
    }
    /**
     * @return {?}
     */
    ngAfterViewInit() {
        this.testMatElementType();
        // Default event 'click' for mat-checkbox and mat-radio-group types
        // And Default event 'blur' for input, select etc.
        /**
         * try using hostlistener here
         * @type {?}
         */
        const event = !!this.validateOnEvent ? this.validateOnEvent : (this.isRadioButton || this.isCheckbox) ? 'click' : 'blur';
        this.renderer2.listen(this.elementRef.nativeElement, event, (/**
         * @return {?}
         */
        () => {
            // todo call central validation runner here
            // Delay the execution of validate by 200ms as the ngModel value needs to be updated.
            // This is workaround for mat-checkbox and mat-radio-group elements
            /** @type {?} */
            const delay = (this.isRadioButton || this.isCheckbox) ? 200 : 0;
            // run validation after the delay
            this.ngZone.runOutsideAngular((/**
             * @return {?}
             */
            () => {
                setTimeout((/**
                 * @return {?}
                 */
                () => {
                    this.validationService.validate(this.policy, this.model, this.validateProperty); // no need to subscribe here
                }), delay);
            }));
        }));
        // this.ngModel.valueChanges.subscribe((value)=>{
        //   this.validationService.validate(this.policy, this.model, this.validateProperty); // no need to subscribe here
        // });
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
    }
    /**
     * @return {?}
     */
    testMatElementType() {
        /** @type {?} */
        const node = this.elementRef.nativeElement;
        this.depthOfParentNode = 5; // till what depth we want to search
        // till what depth we want to search
        /** @type {?} */
        const classList = [...node.classList];
        /** @type {?} */
        const elementClass = classList.reduce((/**
         * @param {?} acc
         * @param {?} value
         * @return {?}
         */
        (acc, value) => {
            value = value.toUpperCase();
            if (value === 'MAT-INPUT-ELEMENT' || value === 'MAT-SELECT') {
                acc.push('MAT-FORM-FIELD');
                return acc;
            }
            else if (value === 'MAT-RADIO-INPUT' || value === 'MAT-RADIO-GROUP') {
                acc.push('MAT-RADIO-GROUP');
                return acc;
            }
            else if (value === 'MAT-CHECKBOX-INPUT' || value === 'MAT-CHECKBOX') {
                acc.push('MAT-CHECKBOX');
                return acc;
            }
            else {
                return acc;
            }
        }), []);
        switch (elementClass[0].toUpperCase()) {
            case 'MAT-CHECKBOX':
                this.parentNodeName = 'MAT-CHECKBOX';
                this.depthOfParentNode = 3;
                this.isCheckbox = true;
                break;
            case 'MAT-RADIO-GROUP':
                this.parentNodeName = 'MAT-RADIO-GROUP';
                this.depthOfParentNode = 3;
                this.isRadioButton = true;
                break;
            case 'MAT-FORM-FIELD':
            case 'MAT-SELECT':
                this.parentNodeName = 'MAT-FORM-FIELD';
                this.depthOfParentNode = 4;
                this.isInput = true;
                break;
            default:
                this.parentNodeName = 'MAT-FORM-FIELD';
                this.depthOfParentNode = 5;
                this.isInput = true;
                break;
        }
        if (this.isCheckbox || this.isRadioButton) {
            this.renderer2.addClass(node, 'mat-form-field');
            /** @type {?} */
            let errorValidationContainer = node.querySelector('.ui-validation-transitionMessages');
            if (errorValidationContainer === null) {
                errorValidationContainer = document.createElement('div');
                errorValidationContainer.className = 'ui-validation-transitionMessages';
                node.appendChild(errorValidationContainer);
            }
        }
    }
    /**
     * @param {?} errorMsg
     * @return {?}
     */
    addErrorToMatControl(errorMsg) {
        /** @type {?} */
        let node = this.elementRef.nativeElement;
        /** @type {?} */
        const maxDept = 5;
        // Using this as mat elements depth of Parent doesn't exceed 5
        /** @type {?} */
        let loopCount = 0;
        while (loopCount < maxDept) {
            if (node.nodeName.toUpperCase() === this.parentNodeName) {
                break;
            }
            else {
                node = node.parentNode;
            }
            loopCount++;
        }
        if (this.isInput) {
            /** @type {?} */
            const matErrorContainer = node.querySelector('.ng-trigger-transitionMessages');
            this.addMatErrorsToHtml(matErrorContainer, node, errorMsg, 'mat-form-field-invalid');
        }
        else if (this.isRadioButton || this.isCheckbox) {
            /** @type {?} */
            const matErrorContainer = node.querySelector('.ui-validation-transitionMessages');
            /** @type {?} */
            const errorClass = this.isCheckbox ? 'mat-checkbox-invalid' : 'mat-radio-invalid';
            this.addMatErrorsToHtml(matErrorContainer, node, errorMsg, errorClass);
        }
    }
    ;
    /**
     * @param {?} errorElement
     * @param {?} node
     * @param {?} errorMsg
     * @param {?} cssErrorClass
     * @return {?}
     */
    addMatErrorsToHtml(errorElement, node, errorMsg, cssErrorClass) {
        if (!!errorElement) {
            /** @type {?} */
            const errorExistingNodes = errorElement.querySelectorAll('mat-error');
            this.renderer2.removeClass(node, cssErrorClass);
            if (!!errorExistingNodes && errorExistingNodes.length > 0) {
                errorExistingNodes.forEach((/**
                 * @param {?} errorNode
                 * @return {?}
                 */
                errorNode => {
                    this.renderer2.removeChild(errorElement, errorNode);
                }));
            }
            // add mat-error block
            if (!!errorMsg) {
                /** @type {?} */
                const errorElem = document.createElement('mat-error');
                /** @type {?} */
                const textNode = document.createTextNode(errorMsg);
                errorElem.appendChild(textNode);
                errorElem.className = 'mat-error';
                /** @type {?} */
                const refDivNode = errorElement.querySelector('div');
                this.renderer2.insertBefore(errorElement, errorElem, refDivNode);
                this.renderer2.addClass(node, cssErrorClass);
            }
        }
    }
    ;
}
NgxMatValidatorDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngModel][ngxMatValidate]',
            },] }
];
/** @nocollapse */
NgxMatValidatorDirective.ctorParameters = () => [
    { type: NgxValidationRunnerService },
    { type: ElementRef },
    { type: NgZone },
    { type: Renderer2 },
    { type: undefined, decorators: [{ type: Inject, args: [WINDOW,] }] }
];
NgxMatValidatorDirective.propDecorators = {
    validateProperty: [{ type: Input }],
    model: [{ type: Input }],
    policy: [{ type: Input }],
    validateOnEvent: [{ type: Input }]
};
if (false) {
    /** @type {?} */
    NgxMatValidatorDirective.prototype.validateProperty;
    /** @type {?} */
    NgxMatValidatorDirective.prototype.model;
    /** @type {?} */
    NgxMatValidatorDirective.prototype.policy;
    /** @type {?} */
    NgxMatValidatorDirective.prototype.validateOnEvent;
    /** @type {?} */
    NgxMatValidatorDirective.prototype.parentNodeName;
    /** @type {?} */
    NgxMatValidatorDirective.prototype.depthOfParentNode;
    /** @type {?} */
    NgxMatValidatorDirective.prototype.isInput;
    /** @type {?} */
    NgxMatValidatorDirective.prototype.isCheckbox;
    /** @type {?} */
    NgxMatValidatorDirective.prototype.isRadioButton;
    /**
     * @type {?}
     * @private
     */
    NgxMatValidatorDirective.prototype.validationService;
    /**
     * @type {?}
     * @private
     */
    NgxMatValidatorDirective.prototype.elementRef;
    /**
     * @type {?}
     * @private
     */
    NgxMatValidatorDirective.prototype.ngZone;
    /**
     * @type {?}
     * @private
     */
    NgxMatValidatorDirective.prototype.renderer2;
    /**
     * @type {?}
     * @private
     */
    NgxMatValidatorDirective.prototype._window;
    /* Skipping unhandled member: ;*/
    /* Skipping unhandled member: ;*/
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/ngx-component-validator.directive.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// Top level component validation enabler
// this shall check for updated validation result and
// publish to ngx-mat-validator via observable and runner service
class NgxComponentValidatorDirective {
    /**
     * @param {?} differs
     * @param {?} ngZone
     * @param {?} validationService
     * @param {?} _window
     */
    constructor(differs, ngZone, validationService, _window) {
        this.differs = differs;
        this.ngZone = ngZone;
        this.validationService = validationService;
        this._window = _window;
        this.trackByFcForNgxValidationResult = (/**
         * @param {?} index
         * @param {?} item
         * @return {?}
         */
        (index, item) => {
            return `${item.propertyName} - ${item.errorMessage}`;
        });
    }
    /**
     * @param {?} changes
     * @return {?}
     */
    ngOnChanges(changes) {
        if (this.policy && this.policyFnRef) {
            /** @type {?} */
            const policy = this.validationService.register(this.policy, new this.policyFnRef());
            if (!this._iterableValidationDiffer) {
                this._iterableValidationDiffer = this.differs.find([]).create(this.trackByFcForNgxValidationResult);
            }
            if (!this._iterableRequiredDiffer) {
                this._iterableRequiredDiffer = this.differs.find([]).create(this.trackByFcForNgxValidationResult);
            }
            // subscribe to observable
            if (!this.validationListenerSubscription)
                this.validationListenerSubscription =
                    this.validationService.getValidationListener(this.policy).subscribe((/**
                     * @param {?} validatorResult
                     * @return {?}
                     */
                    (validatorResult) => {
                        // debugger
                        if (validatorResult.type === 'validation') {
                            /** @type {?} */
                            const differResult = this._iterableValidationDiffer.diff(validatorResult.errors);
                            this.ngZone.runOutsideAngular((/**
                             * @return {?}
                             */
                            () => {
                                if (differResult) {
                                    differResult.forEachRemovedItem((/**
                                     * @param {?} changeRecord
                                     * @return {?}
                                     */
                                    (changeRecord) => {
                                        /** @type {?} */
                                        var payload = Object.assign(Object.assign({}, changeRecord.item), { operation: 'removed', type: 'validation' });
                                        /**
                                         * https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage
                                         */
                                        this._window.postMessage(payload, this._window.location.origin);
                                    }));
                                    differResult.forEachAddedItem((/**
                                     * @param {?} changeRecord
                                     * @return {?}
                                     */
                                    (changeRecord) => {
                                        /** @type {?} */
                                        var payload = Object.assign(Object.assign({}, changeRecord.item), { operation: 'added', type: 'validation' });
                                        /**
                                         * https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage
                                         */
                                        this._window.postMessage(payload, this._window.location.origin);
                                    }));
                                }
                            }));
                        }
                        else {
                            /** do nothing */
                        }
                    }));
            if (!this.requiredListenerSubscription)
                this.requiredListenerSubscription =
                    this.validationService.getRequiredCheckListener(this.policy).subscribe((/**
                     * @param {?} validatorResult
                     * @return {?}
                     */
                    (validatorResult) => {
                        // debugger
                        if (validatorResult.type === 'required') {
                            /** @type {?} */
                            const differResult = this._iterableRequiredDiffer.diff(validatorResult.errors);
                            if (differResult) {
                                this.ngZone.runOutsideAngular((/**
                                 * @return {?}
                                 */
                                () => {
                                    differResult.forEachRemovedItem((/**
                                     * @param {?} changeRecord
                                     * @return {?}
                                     */
                                    (changeRecord) => {
                                        /** @type {?} */
                                        var payload = Object.assign(Object.assign({}, changeRecord.item), { operation: 'removed', type: 'required' });
                                        /**
                                         * https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage
                                         */
                                        this._window.postMessage(payload, this._window.location.origin);
                                    }));
                                    differResult.forEachAddedItem((/**
                                     * @param {?} changeRecord
                                     * @return {?}
                                     */
                                    (changeRecord) => {
                                        /** @type {?} */
                                        var payload = Object.assign(Object.assign({}, changeRecord.item), { operation: 'added', type: 'required' });
                                        /**
                                         * https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage
                                         */
                                        this._window.postMessage(payload, this._window.location.origin);
                                    }));
                                }));
                            }
                        }
                        else {
                            /** do nothing */
                        }
                    }));
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.validationListenerSubscription.unsubscribe();
        this.requiredListenerSubscription.unsubscribe();
    }
}
NgxComponentValidatorDirective.decorators = [
    { type: Directive, args: [{
                selector: 'ngx-component-validator'
            },] }
];
/** @nocollapse */
NgxComponentValidatorDirective.ctorParameters = () => [
    { type: IterableDiffers },
    { type: NgZone },
    { type: NgxValidationRunnerService },
    { type: undefined, decorators: [{ type: Inject, args: [WINDOW,] }] }
];
NgxComponentValidatorDirective.propDecorators = {
    model: [{ type: Input }],
    policy: [{ type: Input }],
    policyFnRef: [{ type: Input }]
};
if (false) {
    /** @type {?} */
    NgxComponentValidatorDirective.prototype.model;
    /** @type {?} */
    NgxComponentValidatorDirective.prototype.policy;
    /** @type {?} */
    NgxComponentValidatorDirective.prototype.policyFnRef;
    /**
     * @type {?}
     * @private
     */
    NgxComponentValidatorDirective.prototype._iterableValidationDiffer;
    /**
     * @type {?}
     * @private
     */
    NgxComponentValidatorDirective.prototype._iterableRequiredDiffer;
    /**
     * @type {?}
     * @private
     */
    NgxComponentValidatorDirective.prototype.validationListenerSubscription;
    /**
     * @type {?}
     * @private
     */
    NgxComponentValidatorDirective.prototype.requiredListenerSubscription;
    /** @type {?} */
    NgxComponentValidatorDirective.prototype.trackByFcForNgxValidationResult;
    /**
     * @type {?}
     * @private
     */
    NgxComponentValidatorDirective.prototype.differs;
    /**
     * @type {?}
     * @private
     */
    NgxComponentValidatorDirective.prototype.ngZone;
    /**
     * @type {?}
     * @private
     */
    NgxComponentValidatorDirective.prototype.validationService;
    /**
     * @type {?}
     * @private
     */
    NgxComponentValidatorDirective.prototype._window;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/ngx-material-required.directive.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// TODO : abstract the validation logic in a base class and use it in a child class for
// any type of UI, e.g. material, bootstrap etc.
// usage    
//   ngxMatRequired [validateProperty]="person.age" [model]="person" [policy]="PERSON_POLICY_NAME" [(ngModel)]="person.age"
class NgxMatRequiredDirective {
    /**
     * @param {?} validationService
     * @param {?} elementRef
     * @param {?} ngZone
     * @param {?} renderer2
     * @param {?} ngModel
     * @param {?} _window
     */
    constructor(validationService, elementRef, ngZone, renderer2, ngModel, _window) {
        this.validationService = validationService;
        this.elementRef = elementRef;
        this.ngZone = ngZone;
        this.renderer2 = renderer2;
        this.ngModel = ngModel;
        this._window = _window;
        this.validateProperty = {};
        this.parentNodeName = '';
        this.isInput = false;
        this.isCheckbox = false;
        this.isRadioButton = false;
        this.addAsteriskToMatControl = (/**
         * @param {?} errorMsg
         * @return {?}
         */
        (errorMsg) => {
            /** @type {?} */
            let node = this.elementRef.nativeElement;
            /** @type {?} */
            const maxDept = 5;
            // mat elements parent's depth doesn't exceed 5
            /** @type {?} */
            let loopCount = 0;
            while (loopCount < maxDept) {
                if (node.nodeName.toUpperCase() === this.parentNodeName) {
                    break;
                }
                else {
                    node = node.parentNode;
                }
                loopCount++;
            }
            if (this.isInput) {
                /** @type {?} */
                const inputFormFieldElement = node.querySelector('.mat-form-field-label');
                if (inputFormFieldElement) {
                    /** @type {?} */
                    const spanRequiredContainer = inputFormFieldElement.querySelectorAll(".mat-placeholder-required");
                    this.addAsteriskToSpanElement(spanRequiredContainer, inputFormFieldElement, errorMsg);
                }
            }
            if (this.isCheckbox) {
                /** @type {?} */
                const checkboxSpanElement = node.querySelector('.mat-checkbox-label');
                if (checkboxSpanElement) {
                    /** @type {?} */
                    const spanRequiredContainer = checkboxSpanElement.querySelectorAll(".mat-placeholder-required");
                    this.addAsteriskToSpanElement(spanRequiredContainer, checkboxSpanElement, checkboxSpanElement);
                }
            }
            if (this.isRadioButton) {
                /** @type {?} */
                const matradioSpanElement = node.querySelectorAll('.mat-radio-label-content');
                if (matradioSpanElement && matradioSpanElement.length > 0) {
                    matradioSpanElement.forEach((/**
                     * @param {?} elem
                     * @return {?}
                     */
                    elem => {
                        /** @type {?} */
                        const spanRequiredContainer = elem.querySelectorAll(".mat-placeholder-required");
                        this.addAsteriskToSpanElement(spanRequiredContainer, elem, errorMsg);
                    }));
                }
            }
        });
        this.addAsteriskToSpanElement = (/**
         * @param {?} spanRequiredContainer
         * @param {?} inputSpanElement
         * @param {?} errorMsg
         * @return {?}
         */
        function (spanRequiredContainer, inputSpanElement, errorMsg) {
            if (spanRequiredContainer && spanRequiredContainer.length > 0) {
                spanRequiredContainer.forEach((/**
                 * @param {?} requiredNode
                 * @return {?}
                 */
                requiredNode => {
                    this.renderer2.removeChild(inputSpanElement, requiredNode);
                }));
            }
            if (!!errorMsg) {
                /** @type {?} */
                const starSpanElement = document.createElement('span');
                /** @type {?} */
                const textNode = document.createTextNode(" *");
                starSpanElement.appendChild(textNode);
                this.renderer2.addClass(starSpanElement, "mat-placeholder-required");
                this.renderer2.addClass(starSpanElement, "mat-form-field-required-marker");
                this.renderer2.addClass(starSpanElement, "label-required");
                inputSpanElement.appendChild(starSpanElement);
                /**
                 * label-required class controls the color of asterisk
                 * add color in demo project styles.css
                 * .label-required { color: red; }
                 */
                // this.renderer2.addClass(inputSpanElement, "label-required");
            }
            else if (inputSpanElement.classList && inputSpanElement.classList.value.indexOf("label-required") > -1) {
                this.renderer2.removeClass(inputSpanElement, "label-required");
            }
        });
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        /**
         * add message event listener
         * message shall come from component validator directive
         */
        // https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage
        this.ngZone.runOutsideAngular((/**
         * @return {?}
         */
        () => {
            // todo : try using https://developer.mozilla.org/en-US/docs/Web/API/Broadcast_Channel_API
            this._window.addEventListener("message", (/**
             * @param {?} event
             * @return {?}
             */
            (event) => {
                if (event.data.type === 'required' && event.data.propertyName === this.validateProperty) {
                    this.addAsteriskToMatControl(event.data.operation === 'added' ? event.data.errorMessage : null);
                }
            }), false);
        }));
    }
    /**
     * @return {?}
     */
    ngAfterViewInit() {
        this.testMatElementType();
        this.ngModelValueChangesSubscription =
            this.ngModel.valueChanges.subscribe((/**
             * @param {?} value
             * @return {?}
             */
            (value) => {
                // delay to update the model
                this.ngZone.runOutsideAngular((/**
                 * @return {?}
                 */
                () => {
                    setTimeout((/**
                     * @return {?}
                     */
                    () => {
                        this.validationService.markForRequiredCheck(this.policy, this.model); // no need to subscribe here
                    }), 200);
                }));
            }));
    }
    /**
     * @param {?} changes
     * @return {?}
     */
    ngOnChanges(changes) {
        // TODO : trigger validation of whole model
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.ngModelValueChangesSubscription.unsubscribe();
    }
    /**
     * @return {?}
     */
    testMatElementType() {
        /** @type {?} */
        const node = this.elementRef.nativeElement;
        this.depthOfParentNode = 5; // till what depth we want to search
        // till what depth we want to search
        /** @type {?} */
        const classList = [...node.classList];
        /** @type {?} */
        const elementClass = classList.reduce((/**
         * @param {?} acc
         * @param {?} value
         * @return {?}
         */
        (acc, value) => {
            value = value.toUpperCase();
            if (value === 'MAT-INPUT-ELEMENT' || value === 'MAT-SELECT') {
                acc.push('MAT-FORM-FIELD');
                return acc;
            }
            else if (value === 'MAT-RADIO-INPUT' || value === 'MAT-RADIO-GROUP') {
                acc.push('MAT-RADIO-GROUP');
                return acc;
            }
            else if (value === 'MAT-CHECKBOX-INPUT' || value === 'MAT-CHECKBOX') {
                acc.push('MAT-CHECKBOX');
                return acc;
            }
            else {
                return acc;
            }
        }), []);
        switch (elementClass[0].toUpperCase()) {
            case 'MAT-CHECKBOX':
                this.parentNodeName = 'MAT-CHECKBOX';
                this.depthOfParentNode = 3;
                this.isCheckbox = true;
                break;
            case 'MAT-RADIO-GROUP':
                this.parentNodeName = 'MAT-RADIO-GROUP';
                this.depthOfParentNode = 3;
                this.isRadioButton = true;
                break;
            case 'MAT-FORM-FIELD':
            case 'MAT-SELECT':
                this.parentNodeName = 'MAT-FORM-FIELD';
                this.depthOfParentNode = 4;
                this.isInput = true;
                break;
            default:
                this.parentNodeName = 'MAT-FORM-FIELD';
                this.depthOfParentNode = 5;
                this.isInput = true;
                break;
        }
        if (this.isCheckbox || this.isRadioButton) {
            this.renderer2.addClass(node, 'mat-form-field');
            /** @type {?} */
            let errorValidationContainer = node.querySelector('.ui-validation-transitionMessages');
            if (errorValidationContainer === null) {
                errorValidationContainer = document.createElement('div');
                errorValidationContainer.className = 'ui-validation-transitionMessages';
                node.appendChild(errorValidationContainer);
            }
        }
    }
}
NgxMatRequiredDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngModel][ngxMatRequired]',
            },] }
];
/** @nocollapse */
NgxMatRequiredDirective.ctorParameters = () => [
    { type: NgxValidationRunnerService },
    { type: ElementRef },
    { type: NgZone },
    { type: Renderer2 },
    { type: NgModel },
    { type: undefined, decorators: [{ type: Inject, args: [WINDOW,] }] }
];
NgxMatRequiredDirective.propDecorators = {
    validateProperty: [{ type: Input }],
    model: [{ type: Input }],
    policy: [{ type: Input }]
};
if (false) {
    /** @type {?} */
    NgxMatRequiredDirective.prototype.validateProperty;
    /** @type {?} */
    NgxMatRequiredDirective.prototype.model;
    /** @type {?} */
    NgxMatRequiredDirective.prototype.policy;
    /** @type {?} */
    NgxMatRequiredDirective.prototype.parentNodeName;
    /** @type {?} */
    NgxMatRequiredDirective.prototype.depthOfParentNode;
    /** @type {?} */
    NgxMatRequiredDirective.prototype.isInput;
    /** @type {?} */
    NgxMatRequiredDirective.prototype.isCheckbox;
    /** @type {?} */
    NgxMatRequiredDirective.prototype.isRadioButton;
    /**
     * @type {?}
     * @private
     */
    NgxMatRequiredDirective.prototype.ngModelValueChangesSubscription;
    /** @type {?} */
    NgxMatRequiredDirective.prototype.addAsteriskToMatControl;
    /** @type {?} */
    NgxMatRequiredDirective.prototype.addAsteriskToSpanElement;
    /**
     * @type {?}
     * @private
     */
    NgxMatRequiredDirective.prototype.validationService;
    /**
     * @type {?}
     * @private
     */
    NgxMatRequiredDirective.prototype.elementRef;
    /**
     * @type {?}
     * @private
     */
    NgxMatRequiredDirective.prototype.ngZone;
    /**
     * @type {?}
     * @private
     */
    NgxMatRequiredDirective.prototype.renderer2;
    /**
     * @type {?}
     * @private
     */
    NgxMatRequiredDirective.prototype.ngModel;
    /**
     * @type {?}
     * @private
     */
    NgxMatRequiredDirective.prototype._window;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/ngx-bootstrap-validator.directive.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// TODO : abstract the validation logic in a base class and use it in a child class for
// any type of UI, e.g. material, bootstrap etc.
// usage    
//   ngxBootStrapValidate [validateProperty]="person.age" [model]="person" [policy]="PERSON_POLICY_NAME" [(ngModel)]="person.age"
// try follow https://github.com/rsaenen/ngx-custom-validators/blob/master/src/app/less-than/directive.ts
class NgxBootStrapValidatorDirective {
    /**
     * @param {?} validationService
     * @param {?} elementRef
     * @param {?} ngZone
     * @param {?} renderer2
     * @param {?} _window
     */
    constructor(validationService, elementRef, ngZone, renderer2, _window) {
        this.validationService = validationService;
        this.elementRef = elementRef;
        this.ngZone = ngZone;
        this.renderer2 = renderer2;
        this._window = _window;
        this.validateProperty = {};
        this.isInput = false;
        this.isCheckbox = false;
        this.isRadioButton = false;
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        /**
         * add message event listener
         * message shall come from component validator directive
         */
        // https://developer.mozilla.org/en-US/docs/Web/API/Window/postMessage
        this._window.addEventListener("message", (/**
         * @param {?} event
         * @return {?}
         */
        (event) => {
            if (event.data.type === 'validation' && event.data.propertyName === this.validateProperty) {
                this.addErrorToBootstrapControl(event.data.operation === 'added' ? event.data.errorMessage : null);
            }
        }), false);
    }
    /**
     * @return {?}
     */
    ngAfterViewInit() {
        this.testBootstrapElementType();
        // Default event 'click' for mat-checkbox and mat-radio-group types
        // And Default event 'blur' for input, select etc.
        /**
         * try using hostlistener here
         * @type {?}
         */
        const event = !!this.validateOnEvent ? this.validateOnEvent : (this.isRadioButton || this.isCheckbox) ? 'click' : 'blur';
        this.renderer2.listen(this.elementRef.nativeElement, event, (/**
         * @return {?}
         */
        () => {
            // todo call central validation runner here
            // Delay the execution of validate by 200ms as the ngModel value needs to be updated.
            // This is workaround for mat-checkbox and mat-radio-group elements
            /** @type {?} */
            const delay = (this.isRadioButton || this.isCheckbox) ? 200 : 0;
            // run validation after the delay
            this.ngZone.runOutsideAngular((/**
             * @return {?}
             */
            () => {
                setTimeout((/**
                 * @return {?}
                 */
                () => {
                    this.validationService.validate(this.policy, this.model, this.validateProperty); // no need to subscribe here
                }), delay);
            }));
        }));
    }
    /**
     * @param {?} changes
     * @return {?}
     */
    ngOnChanges(changes) {
        // TODO : trigger validation of whole model
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
    }
    /**
     * @return {?}
     */
    testBootstrapElementType() {
        /** @type {?} */
        const node = this.elementRef.nativeElement;
        if (node.type === 'text') {
            this.isInput = true;
            this.immediateParentNode = node.parentNode;
            this.formGroupContainerDiv = this.immediateParentNode.parentNode;
        }
        else if (node.type === 'radio') {
            this.isRadioButton = true;
            this.immediateParentNode = node.parentNode.parentNode;
            this.formGroupContainerDiv = this.immediateParentNode.parentNode;
        }
    }
    /**
     * @param {?} errorMsg
     * @return {?}
     */
    addErrorToBootstrapControl(errorMsg) {
        /** @type {?} */
        let node = this.elementRef.nativeElement;
        if (errorMsg) {
            if (this.isRadioButton && this.formGroupContainerDiv.className.indexOf('border') > -1) {
                return;
            }
            this.renderer2.addClass(this.formGroupContainerDiv, 'border');
            this.renderer2.addClass(this.formGroupContainerDiv, 'border-danger');
            /** @type {?} */
            const errorHolder = document.createElement('span');
            /** @type {?} */
            const textNode = document.createTextNode(errorMsg);
            errorHolder.appendChild(textNode);
            this.immediateParentNode.appendChild(errorHolder);
        }
        else {
            this.renderer2.removeClass(this.formGroupContainerDiv, 'border');
            this.renderer2.removeClass(this.formGroupContainerDiv, 'border-danger');
            /** @type {?} */
            const errorTextNode = this.immediateParentNode.lastChild;
            if (errorTextNode.tagName === 'SPAN') {
                this.immediateParentNode.removeChild(errorTextNode);
            }
        }
    }
    ;
}
NgxBootStrapValidatorDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngModel][ngxBootStrapValidate]',
            },] }
];
/** @nocollapse */
NgxBootStrapValidatorDirective.ctorParameters = () => [
    { type: NgxValidationRunnerService },
    { type: ElementRef },
    { type: NgZone },
    { type: Renderer2 },
    { type: undefined, decorators: [{ type: Inject, args: [WINDOW,] }] }
];
NgxBootStrapValidatorDirective.propDecorators = {
    validateProperty: [{ type: Input }],
    model: [{ type: Input }],
    policy: [{ type: Input }],
    validateOnEvent: [{ type: Input }]
};
if (false) {
    /** @type {?} */
    NgxBootStrapValidatorDirective.prototype.validateProperty;
    /** @type {?} */
    NgxBootStrapValidatorDirective.prototype.model;
    /** @type {?} */
    NgxBootStrapValidatorDirective.prototype.policy;
    /** @type {?} */
    NgxBootStrapValidatorDirective.prototype.validateOnEvent;
    /** @type {?} */
    NgxBootStrapValidatorDirective.prototype.immediateParentNode;
    /** @type {?} */
    NgxBootStrapValidatorDirective.prototype.formGroupContainerDiv;
    /** @type {?} */
    NgxBootStrapValidatorDirective.prototype.isInput;
    /** @type {?} */
    NgxBootStrapValidatorDirective.prototype.isCheckbox;
    /** @type {?} */
    NgxBootStrapValidatorDirective.prototype.isRadioButton;
    /**
     * @type {?}
     * @private
     */
    NgxBootStrapValidatorDirective.prototype.validationService;
    /**
     * @type {?}
     * @private
     */
    NgxBootStrapValidatorDirective.prototype.elementRef;
    /**
     * @type {?}
     * @private
     */
    NgxBootStrapValidatorDirective.prototype.ngZone;
    /**
     * @type {?}
     * @private
     */
    NgxBootStrapValidatorDirective.prototype.renderer2;
    /**
     * @type {?}
     * @private
     */
    NgxBootStrapValidatorDirective.prototype._window;
    /* Skipping unhandled member: ;*/
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/ngx-mat-group-validity-check.directive.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
// usage    
// <div ngxBootStrapValidate="group-1" (groupValidationStatusChange)="groupValidationStatusChangeHandler($event)" >
class NgxMatGroupValidityCheckDirective {
    /**
     * @param {?} elementRef
     */
    constructor(elementRef) {
        this.elementRef = elementRef;
        this._requiredStarCount = 0; // goes +ve
        // goes +ve
        this._validationMatErrorCount = 0; // goes -ve
        // goes -ve
        this._ngPristineClassCount = 0; //goes +ve
        //goes +ve
        this._ngValidClassCount = 0; //goes -ve
        this.groupValidationStatusChange = new EventEmitter();
    }
    /**
     * @return {?}
     */
    ngOnInit() { }
    /**
     * @return {?}
     */
    ngAfterViewInit() {
        // Select the node that will be observed for mutations
        /** @type {?} */
        const targetNode = this.elementRef.nativeElement;
        // document.getElementById('myForm');
        // Options for the observer (which mutations to observe)
        /** @type {?} */
        const config = { attributes: true, childList: true, subtree: true };
        // Create an observer instance linked to the callback function
        this.mutationObserver = new MutationObserver((/**
         * @param {?} mutationsList
         * @param {?} observer
         * @return {?}
         */
        (mutationsList, observer) => {
            for (let mutation of mutationsList) {
                if (mutation.type === 'childList') {
                    if (mutation.addedNodes.length === 1 && ((/** @type {?} */ (mutation.addedNodes[0]))).tagName === 'SPAN') {
                        if (((/** @type {?} */ (mutation.addedNodes[0]))).innerHTML.trim() === "*") {
                            this._requiredStarCount++;
                        }
                        else {
                            // something else found
                            // console.error(mutation.addedNodes[0]);
                        }
                    }
                    else if (mutation.addedNodes.length === 1 && ((/** @type {?} */ (mutation.addedNodes[0]))).tagName === 'MAT-ERROR') {
                        // this._validationMatErrorCount++;
                    }
                    else if (mutation.removedNodes.length === 1 && ((/** @type {?} */ (mutation.removedNodes[0]))).tagName === 'SPAN') {
                        if (((/** @type {?} */ (mutation.removedNodes[0]))).innerHTML.trim() === "*") {
                            this._requiredStarCount--;
                        }
                    }
                    else if (mutation.removedNodes.length === 1 && ((/** @type {?} */ (mutation.removedNodes[0]))).tagName === 'MAT-ERROR') {
                        this._validationMatErrorCount--;
                    }
                    // console.log(`Required Star Count >> ${this._requiredStarCount}`);
                }
                else if (mutation.type === 'attributes' && mutation.attributeName === 'class' && ((/** @type {?} */ (mutation.target))).tagName === 'MAT-FORM-FIELD') {
                    /** @type {?} */
                    const classList = ((/** @type {?} */ (mutation.target))).classList;
                    if (classList.contains('ng-untouched') && !classList.contains('mat-focused')) {
                        this._ngPristineClassCount++;
                    }
                    else if (classList.contains('mat-form-field-invalid') && !classList.contains('mat-focused')) {
                        this._ngValidClassCount--;
                    }
                    else if (classList.contains('ng-dirty') && !classList.contains('mat-focused') && !classList.contains('mat-form-field-invalid')) {
                        this._ngValidClassCount = this.elementRef.nativeElement.querySelectorAll('mat-form-field.ng-dirty').length;
                        // this._ngValidClassCount++;
                    }
                    // console.log('The ' + mutation.attributeName + ' attribute was modified.');
                }
            }
            // check for any error left
            if ((this._requiredStarCount + this._validationMatErrorCount === 0) || (this._ngPristineClassCount + this._validationMatErrorCount === 0) ||
                (this._ngPristineClassCount === this._ngValidClassCount)) {
                console.info('No validation error');
                this.groupValidationStatusChange.emit({ section: this.groupName || 'section', isValid: 'valid' });
            }
            else {
                console.info('Validation error found');
                this.groupValidationStatusChange.emit({ section: this.groupName || 'section', isValid: 'invalid' });
            }
        }));
        // Start observing the target node for configured mutations
        this.mutationObserver.observe(targetNode, config);
    }
    /**
     * @param {?} changes
     * @return {?}
     */
    ngOnChanges(changes) { }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.mutationObserver.disconnect();
    }
}
NgxMatGroupValidityCheckDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngxMatGroupValidityCheck]'
            },] }
];
/** @nocollapse */
NgxMatGroupValidityCheckDirective.ctorParameters = () => [
    { type: ElementRef }
];
NgxMatGroupValidityCheckDirective.propDecorators = {
    groupName: [{ type: Input, args: ['ngxMatGroupValidityCheck',] }],
    groupValidationStatusChange: [{ type: Output }]
};
if (false) {
    /**
     * @type {?}
     * @private
     */
    NgxMatGroupValidityCheckDirective.prototype.mutationObserver;
    /**
     * @type {?}
     * @private
     */
    NgxMatGroupValidityCheckDirective.prototype._requiredStarCount;
    /**
     * @type {?}
     * @private
     */
    NgxMatGroupValidityCheckDirective.prototype._validationMatErrorCount;
    /**
     * @type {?}
     * @private
     */
    NgxMatGroupValidityCheckDirective.prototype._ngPristineClassCount;
    /**
     * @type {?}
     * @private
     */
    NgxMatGroupValidityCheckDirective.prototype._ngValidClassCount;
    /** @type {?} */
    NgxMatGroupValidityCheckDirective.prototype.groupName;
    /** @type {?} */
    NgxMatGroupValidityCheckDirective.prototype.groupValidationStatusChange;
    /**
     * @type {?}
     * @private
     */
    NgxMatGroupValidityCheckDirective.prototype.elementRef;
}

/**
 * @fileoverview added by tsickle
 * Generated from: lib/module.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxValidationModule {
}
NgxValidationModule.decorators = [
    { type: NgModule, args: [{
                exports: [NgxMatValidatorDirective, NgxComponentValidatorDirective, NgxMatRequiredDirective, NgxBootStrapValidatorDirective, NgxMatGroupValidityCheckDirective],
                declarations: [NgxMatValidatorDirective, NgxComponentValidatorDirective, NgxMatRequiredDirective, NgxBootStrapValidatorDirective, NgxMatGroupValidityCheckDirective],
                providers: [NgxValidationRunnerService]
            },] }
];

/**
 * @fileoverview added by tsickle
 * Generated from: public-api.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * Generated from: ngx-model-validator.ts
 * @suppress {checkTypes,constantProperty,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { NgxBootStrapValidatorDirective, NgxComponentValidatorDirective, NgxMatGroupValidityCheckDirective, NgxMatRequiredDirective, NgxMatValidatorDirective, NgxValidationModule, NgxValidationRunnerService, NgxValidatorHelper, WINDOW as ɵa };
//# sourceMappingURL=ngx-model-validator.js.map
